/*
* This file is part of the LibreOffice project.
*
* This Source Code Form is subject to the terms of the Mozilla Public
* License, v. 2.0. If a copy of the MPL was not distributed with this
* file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
#include "Common.h"
#include <cstring>

namespace glTF
{

int gltf_get_file_index_by_name(glTFHandle* handle, const char* name)
{
    if (handle == NULL)
    {
        return -1;
    }
    for (unsigned int i = 0; i < handle->size; i++)
    {
        if ((!strcmp(name, (handle->files[i]->filename))))
        {
            return (int)i;
        }
    }
    return -1;
}

/* -- Attribute -- */
Attribute::Attribute()
    : pData(0)
    , mCount(0)
    , mDataType(DataType_UNKNOW)
{
}

Attribute::~Attribute()
{
    delete[] pData;
}

void Attribute::setDataCount(unsigned int count)
{
    this->mCount = count;
}

unsigned int Attribute::getDataCount() const
{
    return this->mCount;
}

void Attribute::setByteStride(unsigned int byteStride)
{
    this->mByteStride = byteStride;
}

unsigned int Attribute::getByteStride() const
{
    return this->mByteStride;
}

void Attribute::setDataType(DataType_E type)
{
    this->mDataType = type;
}

DataType_E Attribute::getDataType() const
{
    return this->mDataType;
}

void Attribute::setAttributeData(const char* pValue, unsigned int length)
{
    if (0 == pData)
    {
        pData = new char[length];
    }
    memcpy(this->pData, pValue, length);
}

const char* Attribute::getAttributeData() const
{
    return pData;
}

/* -- Light -- */
Light::Light()
    : mPosition(sqrt(2.0f) / 2, -sqrt(2.0f) / 2, 0)
    , mDirection(sqrt(2.0f) / 2, -sqrt(2.0f) / 2, 0)
    , mColor(1.0f, 1.0f, 1.0f)
    , mName("light0Color")
    , mPosFlag(false)
{
    mColorAmbient = new GLfloat[4];
    mColorDiffuse = new GLfloat[4];
    mColorSpecular = new GLfloat[4];
    mPositions = new GLfloat[4];
    for (int i = 0; i < 4; i++)
    {
        mColorAmbient[i]  = 1.0;
        mColorDiffuse[i]  = 1.0;
        mColorSpecular[i] = 1.0;
        mPositions[i]     = 1.0;
    }
}

Light::~Light()
{
    delete[] mColorAmbient;
    delete[] mColorDiffuse;
    delete[] mColorSpecular;
    delete[] mPositions;
}

void Light::setName(std::string Name)
{
    this->mName = Name;
}

const std::string& Light::getName()
{
    return mName;
}

void Light::setType(LightSourceType_E Type)
{
    this->mType = Type;
}

const LightSourceType_E& Light::getLightSourceType()
{
    return mType;
}

void Light::setPosition(const glm::vec3& Position)
{
    this->mPosition = Position;
}

const glm::vec3& Light::getPosition()
{
    return mPosition;
}

void Light::setDirection(glm::vec3 Direction)
{
    this->mDirection = Direction;
}

const glm::vec3& Light::getDirection()
{
    return mDirection;
}

void Light::setAttenuationConstant(float AttenuationConstant)
{
    this->mAttenuationConstant = AttenuationConstant;
}

const float Light::getAttenuationConstant()
{
    return mAttenuationConstant;
}

void Light::setAttenuationLinear(float AttenuationLinear)
{
    this->mAttenuationLinear = AttenuationLinear;
}

const float Light::getAttenuationLinear()
{
    return mAttenuationLinear;
}

void Light::setAttenuationQuadratic(float AttenuationQuadratic)
{
    this->mAttenuationQuadratic = AttenuationQuadratic;
}

const float Light::getAttenuationQuadratic()
{
    return mAttenuationQuadratic;
}

void Light::setColor(glm::vec3 Color)
{
    this->mColor = Color;
}

const glm::vec3 Light::getColor()
{
    return this->mColor;
}

void Light::setColorDiffuse(GLfloat* ColorDiffuse)
{
    for (int i = 0; i < 3; i++)
    {
        mColorDiffuse[i] = *ColorDiffuse;
    }
}

const GLfloat* Light::getColorDiffuse()
{
    return mColorDiffuse;
}

void Light::setColorSpecular(GLfloat* ColorSpecular)
{
    for (int i = 0; i < 3; i++)
    {
        mColorSpecular[i] = *ColorSpecular;
    }
}

const GLfloat* Light::getColorSpecular()
{
    return mColorSpecular;
}

void Light::setColorAmbient(const GLfloat* ColorAmbient)
{
    for (int i = 0; i < 3; i++)
    {
        mColorAmbient[i] = *ColorAmbient;
    }
}

const GLfloat* Light::getColorAmbient()
{
    return mColorAmbient;
}

void Light::setPositions(const GLfloat* position)
{
    memcpy(mPositions, position, sizeof(mPositions) * 4);
}
const GLfloat* Light::getPositions()
{
    return mPositions;
}

void Light::setPosFlag()
{
    mPosFlag = true;
}
const bool Light::getPosFlag()
{
    return mPosFlag;
}

/* -- Camera -- */
ParseCamera::ParseCamera()
    : mXFov(1.0f)
    , mYFov(1.2f)
    , mNear(0.1f)
    , mFar(1000.f)
    , mAspectRatio(0.f)
{
}

ParseCamera::~ParseCamera()
{
}

void ParseCamera::setXFov(float xFov)
{
    this->mXFov = xFov;
}

float ParseCamera::getXFov()
{
    return this->mXFov;
}

void ParseCamera::setYFov(float yFov)
{
    this->mYFov = yFov;
}

float ParseCamera::getYFov()
{
    return this->mYFov;
}

void ParseCamera::setFar(float far)
{
    this->mFar = far;
}

float ParseCamera::getFar()
{
    return this->mFar;
}

void ParseCamera::setNear(float near)
{
    this->mNear = near;
}

float ParseCamera::getNear()
{
    return this->mNear;
}

void ParseCamera::setAspectRatio(float aspectRatio)
{
    this->mAspectRatio = aspectRatio;
}

float ParseCamera::getAspectRatio()
{
    return this->mAspectRatio;
}

/* -- TechAttribute -- */
TechAttribute::TechAttribute()
{
}

TechAttribute::~TechAttribute()
{
}

void TechAttribute::setAttributeName(const std::string& name)
{
    this->mAttributeName = name;
}

const std::string TechAttribute::getAttributeName()
{
    return this->mAttributeName;
}

void TechAttribute::setAttributeIndex(const std::string& index)
{
    this->mAttributeIndex = index;
}

const std::string TechAttribute::getAttributeIndex()
{
    return this->mAttributeIndex;
}

/* -- TechUniform -- */
TechUniform::TechUniform()
{
}

TechUniform::~TechUniform()
{
}

void TechUniform::setUniformName(const std::string& name)
{
    this->mUniformName = name;
}

const std::string& TechUniform::getUniformName()
{
    return this->mUniformName;
}

void TechUniform::setUniformIndex(const std::string& index)
{
    this->mUniformIndex = index;
}

const std::string& TechUniform::getUniformIndex()
{
    return this->mUniformIndex;
}

/* -- Technique -- */
Technique::Technique()
    : mProgramId(0)
    , mTechniqueState(false)
{
    pState = new TechniqueState();
}

Technique::~Technique()
{
    std::vector<TechUniform*>::iterator vit;
    for (vit = mTechUnifVec.begin(); vit != mTechUnifVec.end(); ++vit)
    {
        delete(*vit);
    }

    std::map<std::string, TechAttribute*>::iterator mit;
    for (mit = mTechAttrMap.begin(); mit != mTechAttrMap.end(); ++mit)
    {
        delete mit->second;
    }
    mTechAttrMap.clear();
    freeTechnique();
    delete pState;
}

void Technique::setProgramState(bool flag)
{
    this->mTechniqueState = flag;
}

bool Technique::getProgramState() const
{
    return this->mTechniqueState;
}

void Technique::setProgramId(unsigned int id)
{
    this->mProgramId = id;
}

unsigned int Technique::getProgramId() const
{
    return this->mProgramId;
}

void Technique::setTechId(std::string techId)
{
    this->mTechniqueId = techId;
}

const std::string& Technique::getTechId()
{
    return this->mTechniqueId;
}

void Technique::setVertexShader(const std::string& shaderName)
{
    this->mVShaderName = shaderName;
}

const std::string& Technique::getVertexShader()
{
    return this->mVShaderName;
}

void Technique::setFragmentShader(const std::string& shaderName)
{
    this->mFShaderName = shaderName;
}

const std::string& Technique::getFragmentShader()
{
    return this->mFShaderName;
}

void Technique::insertTechAttribute(const std::string& key,
                                    TechAttribute* pTechAttr)
{
    this->mTechAttrMap.insert(
        std::map<std::string, TechAttribute*>::value_type(key, pTechAttr));
}

TechAttribute* Technique::getTechAttribute(const std::string& key)
{
    std::map<std::string, TechAttribute*>::const_iterator it;
    it = mTechAttrMap.find(key);
    if (it != mTechAttrMap.end())
    {
        return it->second;
    }
    return 0;
}

unsigned int Technique::getTechAttributeSize() const
{
    return this->mTechAttrMap.size();
}

void Technique::pushTechUniform(TechUniform* pUniform)
{
    (this->mTechUnifVec).push_back(pUniform);
}

TechUniform* Technique::getTechUniform(unsigned int index)
{
    if (this->mTechUnifVec.size() <= index)
    {
        return 0;
    }
    return this->mTechUnifVec[index];
}

unsigned int Technique::getTechUniformSize() const
{
    return this->mTechUnifVec.size();
}

TechniqueState* Technique::getTechState()
{
    return pState;
}

int Technique::initTechnique(glTFHandle* pGltfHandle)
{
    int vShaderIdx = gltf_get_file_index_by_name(pGltfHandle,
                     mVShaderName.c_str());
    int fShaderIdx = gltf_get_file_index_by_name(pGltfHandle,
                     mFShaderName.c_str());
    if (vShaderIdx >= 0 && fShaderIdx >= 0)
    {
        const char* pvShader = pGltfHandle->files[vShaderIdx]->buffer;
        size_t ivShaderSize = pGltfHandle->files[vShaderIdx]->size;
        const char* pfShader = pGltfHandle->files[fShaderIdx]->buffer;
        size_t ifShaderSize = pGltfHandle->files[fShaderIdx]->size;
        mProgramId = mShaderProg.createProgram(pvShader, ivShaderSize, pfShader, ifShaderSize);
    }
    if (0 != mProgramId)
    {
        this->setProgramState(true);
        return 0;
    }
    else
    {
        return SHADER_ERROR;
    }
}

void Technique::useTechnique()
{
    mShaderProg.useProgram(0);
    mShaderProg.useProgram(mProgramId);
}

void Technique::freeTechnique()
{
    mShaderProg.deleteProgram(mProgramId);
    this->setProgramState(false);
}

/* -- MaterialProperty -- */
MaterialProperty::MaterialProperty()
    : mTextureType(TextureType_NONE)
    , mDataType(DataType_UNKNOW)
    , mDataLength(0)
    , pData(0)
{
}

MaterialProperty::~MaterialProperty()
{
    delete [] pData;
}

void MaterialProperty::setPropertyName(const std::string& name)
{
    this->mPropertyName = name;
}

const std::string& MaterialProperty::getPropertyName() const
{
    return this->mPropertyName;
}

void MaterialProperty::setImagePath(const std::string& index)
{
    this->mImagePath = index;
}

const std::string& MaterialProperty::getImagePath() const
{
    return this->mImagePath;
}

void MaterialProperty::setDataLength(unsigned int length)
{
    this->mDataLength = length;
}

unsigned int MaterialProperty::getDataLength()
{
    return this->mDataLength;
}

void MaterialProperty::setDataType(DataType_E type)
{
    this->mDataType = type;
}

DataType_E MaterialProperty::getDataType() const
{
    return this->mDataType;
}

void MaterialProperty::setPropertyData(char* pValue, unsigned int size)
{
    if (0 != this->pData)
        delete [] this->pData;
    this->pData = new char[size];
    memcpy(this->pData, pValue, size);
}

const char* MaterialProperty::getPropertyData() const
{
    return pData;
}

/* -- Material -- */
Material::Material()
{
}

Material::~Material()
{
    std::vector<MaterialProperty*>::iterator it;
    for (it = mPropertyVec.begin(); it != mPropertyVec.end(); ++it)
    {
        delete(*it);
    }
    mPropertyVec.clear();
}

void Material::pushMaterialProper(MaterialProperty* pProperty)
{
    this->mPropertyVec.push_back(pProperty);
}

const MaterialProperty* Material::getMaterialProper(unsigned int index)
{
    if (this->mPropertyVec.size() <= index)
    {
        return 0;
    }
    return this->mPropertyVec[index];
}

unsigned int Material::getMaterialProperSize()
{
    return this->mPropertyVec.size();
}

void Material::setTechniqueId(const std::string& techniqueId)
{
    this->mTechniqueId = techniqueId;
}

const std::string& Material::getTechniqueId() const
{
    return this->mTechniqueId;
}

/* -- Primitives -- */
Primitives::Primitives()
{
};

Primitives::~Primitives()
{
    mAttributeMap.clear();
};

void Primitives::setMaterialIndex(std::string materialIndex)
{
    this->mMaterialIndex = materialIndex;
}

const std::string Primitives::getMaterialIndex() const
{
    return this->mMaterialIndex;
}

void Primitives::setIndicesIndex(std::string indicesIndex)
{
    this->mIndicesIndex = indicesIndex;
}

const std::string Primitives::getIndicesIndex() const
{
    return this->mIndicesIndex;
}

void Primitives::insertAttribute(const std::string& key, const std::string& value)
{
    mAttributeMap.insert(
        std::map<std::string, std::string>::value_type(key, value));
}

const std::string Primitives::getAttributeIndex(const std::string& key) const
{
    std::map<std::string, std::string>::const_iterator it;
    it = mAttributeMap.find(key);
    if (it != mAttributeMap.end())
    {
        return it->second;
    }

    return "";
}

unsigned int Primitives::getAttributeSize() const
{
    return this->mAttributeMap.size();
}

/* -- Mesh -- */
Mesh::Mesh()
{
}

Mesh::~Mesh()
{
    std::vector<Primitives*>::iterator it;
    for (it = mPrimitiveVec.begin(); it != mPrimitiveVec.end(); ++it)
    {
        delete(*it);
    }
    mPrimitiveVec.clear();
}

void Mesh::setMeshName(const std::string& mName)
{
    this->mName = mName;
}

const std::string Mesh::getMeshName()
{
    return this->mName;
}

void Mesh::setPrimitiveVec(Primitives* pPrim)
{
    (this->mPrimitiveVec).push_back(pPrim);
}

const Primitives* Mesh::getPrimitiveVec(unsigned int index) const
{
    if (this->mPrimitiveVec.size() <= index)
    {
        return 0;
    }
    return this->mPrimitiveVec[index];
}

unsigned int Mesh::getPrimitiveVecSize() const
{
    return this->mPrimitiveVec.size();
}

/* -- Node -- */
Node::Node()
    : mNodeType(NodeType_Init)
    , pParent(0)
    , pMatrix(0)
    , LightFlag(true)
    , bFlagCamera(false)
{
}

Node::~Node()
{
    std::vector<Node*>::iterator it;
    for (it = mChildrenVec.begin(); it != mChildrenVec.end(); ++it)
    {
        delete(*it);
    }
    mChildrenVec.clear();
    delete pMatrix;
}

void Node::setMatrix(glm::mat4* matrix)
{
    this->pMatrix = matrix;
}

glm::mat4* Node::getMatrix() const
{
    return this->pMatrix;
}

void Node::setCurrentNodeMatrix(const glm::mat4 &currentMatrix)
{
    mCurrentNodeMatrix = currentMatrix;
}

glm::mat4 Node::getCurrentNodeMatrix()
{
    return mCurrentNodeMatrix;
}

void Node::setParentMatrixes()
{
    Node* pCurrentNode = pParent;
    while (pCurrentNode != NULL)
    {
        mParentMatrixes.push_back(pCurrentNode->mCurrentNodeMatrix);
        pCurrentNode = pCurrentNode->pParent;
    }
}
vector<glm::mat4>* Node::getParentMatrixes()
{
    return &mParentMatrixes;
}

void Node::setNodeType(NodeType_E _type)
{
    this->mNodeType = _type;
}

NodeType_E Node::getNodeType() const
{
    return this->mNodeType;
}

void Node::setNodeName(std::string name)
{
    this->mName = name;
}

const std::string Node::getNodeName() const
{
    return this->mName;
}

void Node::setLightIndex(const std::string &lightIndex)
{
    this->mLightIndex = lightIndex;
}

const std::string Node::getLightIndex() const
{
    return this->mLightIndex;
}

void Node::setCameraIndex(std::string cameraIndex)
{
    this->mCameraIndex = cameraIndex;
}

const std::string Node::getCameraIndex()
{
    return this->mCameraIndex;
}

void Node::pushMeshIndex(const std::string& _index)
{
    (this->mMesheIndexVec).push_back(_index);
}

const std::string Node::getMeshIndex(unsigned int index) const
{
    if (this->mMesheIndexVec.size() <= index)
    {
        return NULL;
    }
    return this->mMesheIndexVec[index];
}

unsigned int Node::getMeshIndexSize() const
{
    return this->mMesheIndexVec.size();
}

void Node::pushChildNode(Node* pNode)
{
    (this->mChildrenVec).push_back(pNode);
}

const Node* Node::getChildNode(unsigned int index) const
{
    if (this->mChildrenVec.size() <= index)
    {
        return 0;
    }
    return this->mChildrenVec[index];
}

Node* Node::getChildrenNode(unsigned int index)
{
    if (this->mChildrenVec.size() <= index)
    {
        return 0;
    }
    return this->mChildrenVec[index];
}

unsigned int Node::getChildNodeSize() const
{
    return this->mChildrenVec.size();
}

void Node::setParentNode(Node* pParentNode)
{
    this->pParent = pParentNode;
}

const Node* Node::getParentNode()
{
    return this->pParent;
}

const bool Node::getLightFlag()
{
    return this->LightFlag;
}

void Node::setLightFlag()
{
    this->LightFlag = false;
}



/* -- AnimChannel -- */
AnimChannel::AnimChannel()
    : mCount(0)
    , mMaxTime(0)
    , mMinTime(0)
{
}

AnimChannel::~AnimChannel()
{
    std::vector<QuatKey4>::iterator it4;
    for (it4 = mTimeValue4Vec.begin(); it4 != mTimeValue4Vec.end(); ++it4)
    {
        delete it4->mValue;
    }
    mTimeValue4Vec.clear();

    std::vector<QuatKey3>::iterator it3;
    for (it3 = mTimeValue3Vec.begin(); it3 != mTimeValue3Vec.end(); ++it3)
    {
        delete it3->mValue;
    }
    mTimeValue3Vec.clear();
}

void AnimChannel::setCount(unsigned int cout)
{
    this->mCount = cout;
}

unsigned int AnimChannel::getCount()
{
    return this->mCount;
}

void AnimChannel::pushTimeValue4(double time, glm::vec4* value)
{
    QuatKey4 quatKey;
    quatKey.mTime = time;
    quatKey.mValue = value;
    this->mTimeValue4Vec.push_back(quatKey);
}

glm::vec4* AnimChannel::getTimeValue4(double time)
{
    unsigned int index;
    unsigned int size = mTimeValue4Vec.size();
    for (index = 0; index < size; ++index)
    {
        if (time <= mTimeValue4Vec[index].mTime)
        {
            break;
        }
    }
    return mTimeValue4Vec[index].mValue;
}

void AnimChannel::pushTimeValue3(double time, glm::vec3* value)
{
    QuatKey3 quatKey;
    quatKey.mTime = time;
    quatKey.mValue = value;
    this->mTimeValue3Vec.push_back(quatKey);
}

glm::vec3* AnimChannel::getTimeValue3(double time)
{
    unsigned int index;
    unsigned int size = mTimeValue3Vec.size();
    for (index = 0; index < size; ++index)
    {
        if (time <= mTimeValue3Vec[index].mTime)
        {
            break;
        }
    }
    return mTimeValue3Vec[index].mValue;
}

void AnimChannel::setMaxTime(double time)
{
    this->mMaxTime = time;
}

double AnimChannel::getMaxTime()
{
    return this->mMaxTime;
}

void AnimChannel::setMinTime(double time)
{
    this->mMinTime = time;
}

double AnimChannel::getMinTime()
{
    return this->mMinTime;
}



/* -- Animation -- */
Animation::Animation()
    : mDuration(0)
    , mTicksPerSecond(0)
    , pRotateChl(0)
    , pScaleChl(0)
    , pTransChl(0)
{
}

Animation::~Animation()
{
    delete pRotateChl;
    delete pScaleChl;
    delete pTransChl;
}

void Animation::setDuration(double duration)
{
    this->mDuration = duration;
}

double Animation::getDuration()
{
    return this->mDuration;
}

void Animation::setTicksPerSecond(double ticks)
{
    this->mTicksPerSecond = ticks;
}

double Animation::getTicksPerSecond()
{
    return this->mTicksPerSecond;
}

void Animation::setRotateChl(AnimChannel* channel)
{
    this->pRotateChl = channel;
}

AnimChannel* Animation::getRotateChl()
{
    return this->pRotateChl;
}

void Animation::setScaleChl(AnimChannel* channel)
{
    this->pScaleChl = channel;
}

AnimChannel* Animation::getScaleChl()
{
    return this->pScaleChl;
}

void Animation::setTransChl(AnimChannel* channel)
{
    this->pTransChl = channel;
}

AnimChannel* Animation::getTransChl()
{
    return this->pTransChl;
}

void Animation::setBoneId(std::string boneId)
{
    this->mBoneId = boneId;
}

const std::string& Animation::getBoneId()
{
    return this->mBoneId;
}



/* -- Skin -- */
Skin::Skin()
    : pBindMatrix(0)
{
}

Skin::~Skin()
{
    delete[] pBindMatrix;
    mBoneIdVec.clear();
}

void Skin::setBindMatrix(glm::mat4* matrix)
{
    this->pBindMatrix = matrix;
}

glm::mat4* Skin::getBindMatrix()
{
    return this->pBindMatrix;
}

void Skin::setBindMatrixCount(unsigned int count)
{
    this->mBindMatrixCount = count;
}

unsigned int Skin::getBindMatrixCount()
{
    return this->mBindMatrixCount;
}

void Skin::pushBoneId(std::string boneId)
{
    this->mBoneIdVec.push_back(boneId);
}

const std::string Skin::getBoneId(unsigned int index)
{
    if (this->mBoneIdVec.size() <= index)
    {
        return 0;
    }
    return this->mBoneIdVec[index];
}

unsigned int Skin::getBoneIdSize()
{
    return this->mBoneIdVec.size();
}



/* -- Bone -- */
Bone::Bone()
    : bJointFlag(false)
    , pParent(0)
{
}

Bone::~Bone()
{
    std::vector<Bone*>::iterator it;
    for (it = mChildrenVec.begin(); it != mChildrenVec.end(); ++it)
    {
        delete *it;
    }
    mChildrenVec.clear();
}

void Bone::pushChildBone(Bone* pBone)
{
    this->mChildrenVec.push_back(pBone);
}

Bone* Bone::getChildBone(unsigned int index)
{
    if (this->mChildrenVec.size() <= index)
    {
        return 0;
    }
    return this->mChildrenVec[index];
}

unsigned int Bone::getChildBoneSize()
{
    return mChildrenVec.size();
}

void Bone::setParentBone(Bone* parent)
{
    this->pParent = parent;
}

Bone* Bone::getParentBone()
{
    return this->pParent;
}

void Bone::setJointFlag(bool flag)
{
    this->bJointFlag = flag;
}

bool Bone::getJointFlag()
{
    return this->bJointFlag;
}

void Bone::setNodeName(const std::string& name)
{
    this->mNodeName = name;
}

const std::string& Bone::getNodeName()
{
    return this->mNodeName;
}

void Bone::setJointId(const std::string& jointId)
{
    this->mJointId = jointId;
}

const std::string& Bone::getJointId()
{
    return this->mJointId;
}

void Bone::setGlobleMatrix(glm::mat4* matrix)
{
    memcpy(&mGlobleMatrix, matrix, sizeof(glm::mat4));
}

void Bone::setGlobleMatrix(float* matrix)
{
    memcpy(&mGlobleMatrix, matrix, sizeof(glm::mat4));
}

glm::mat4& Bone::getGlobleMatrix()
{
    return this->mGlobleMatrix;
}

void Bone::setLocalMatrix(const glm::mat4* matrix)
{
    memcpy(&mLocalMatrix, matrix, sizeof(glm::mat4));
}

void Bone::setLocalMatrix(const float* matrix)
{
    memcpy(&mLocalMatrix, matrix, sizeof(glm::mat4));
}

glm::mat4& Bone::getLocalMatrix()
{
    return this->mLocalMatrix;
}

void Bone::setRotate(const glm::vec4* protate)
{
    memcpy(&mRotate, protate, sizeof(glm::vec4));
}

void Bone::setRotate(const float* protate)
{
    memcpy(&mRotate, protate, sizeof(glm::vec4));
}

glm::vec4& Bone::getRotate()
{
    return this->mRotate;
}

void Bone::setScale(const glm::vec3* pscale)
{
    memcpy(&mScale, pscale, sizeof(glm::vec3));
}
void Bone::setScale(const float* pscale)
{
    memcpy(&mScale, pscale, sizeof(glm::vec3));
}

glm::vec3& Bone::getScale()
{
    return this->mScale;
}

void Bone::setTranslate(const glm::vec3* ptrans)
{
    memcpy(&mTranslate, ptrans, sizeof(glm::vec3));
}

void Bone::setTranslate(const float* ptrans)
{
    memcpy(&mTranslate, ptrans, sizeof(glm::vec3));
}

glm::vec3& Bone::getTranslate()
{
    return this->mTranslate;
}



/* -- Scene -- */
Scene::Scene()
    : pRootNode(0)
    , pRootBone(0)
    , pBuffer(0)
    , pGltfHandle(0)
    , pSkin(0)
    , mVertexMin(1e10f, 1e10f, 1e10f)
    , mVertexMax(-1e10f, -1e10f, -1e10f)
{
}

Scene::~Scene()
{
    freeMap(mMeshMap);
    freeMap(mMaterialMap);
    freeMap(mLightMap);
    freeMap(mCameraMap);
    freeMap(mAttributeMap);
    freeMap(mTexturesMap);
    delete[] pBuffer;
    delete pSkin;
    delete pRootBone;
    delete pRootNode;
    mNodeVec.clear();

    std::vector<Technique*>::iterator it;
    for (it = mTechniqueVec.begin(); it != mTechniqueVec.end(); ++it)
    {
        delete *it;
    }
    mTechniqueVec.clear();
}

template<typename M> void Scene::freeMap(M& aMap)
{
    for (typename M::iterator it = aMap.begin(); it != aMap.end(); ++it)
    {
        delete it->second;
    }
    aMap.clear();
}

Mesh* Scene::findMesh(const std::string& key)
{
    std::map<std::string, Mesh*>::iterator it;
    it = mMeshMap.find(key);
    if (it != mMeshMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertMeshMap(const std::string& key, Mesh* pMesh)
{
    this->mMeshMap.insert(
        std::map<std::string, Mesh*>::value_type(key, pMesh));
}

Material* Scene::findMaterial(const std::string& key)
{
    std::map<std::string, Material*>::iterator it;
    it = mMaterialMap.find(key);
    if (it != mMaterialMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertMaterialMap(const std::string& key, Material* pMaterial)
{
    this->mMaterialMap.insert(
        std::map<std::string, Material*>::value_type(key, pMaterial));
}

ParseCamera* Scene::findCamera(const std::string& key)
{
    std::map<std::string, ParseCamera*>::iterator it;
    it = mCameraMap.find(key);
    if (it != mCameraMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertCameraMap(const std::string& key, ParseCamera* pCamera)
{
    this->mCameraMap.insert(
        std::map<std::string, ParseCamera*>::value_type(key, pCamera));
}

const std::string Scene::getCameraIndex() const
{
    return this->pRootNode->getChildrenNode(1)->getCameraIndex();
}

const glm::mat4* Scene::getCameraMatrix() const
{
    return this->pRootNode->getChildrenNode(1)->getMatrix();
}

std::string Scene::GetLightIndex()
{
    for (unsigned int i = 0; i < pRootNode->getChildNodeSize(); i++)
    {
        const Node* pNode = pRootNode->getChildNode(i);
        if (pNode->getLightIndex() != "")
        {
            return pNode->getLightIndex();
        }
    }
    return "";
}

bool Scene::getLight()
{
    return pRootNode->getLightFlag();
}

Light* Scene::findLight(const std::string& key)
{
    std::map<std::string, Light*>::iterator it;
    it = mLightMap.find(key);
    if (it != mLightMap.end())
    {
        return it->second;
    }
    return 0;
}

const Attribute* Scene::findAttribute(const std::string& key)
{
    std::map<std::string, Attribute*>::iterator it;
    it = mAttributeMap.find(key);
    if (it != mAttributeMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertAttributeMap(std::string key, Attribute* pAttribute)
{
    this->mAttributeMap.insert(
        std::map<std::string, Attribute*>::value_type(key, pAttribute));
}

glm::mat4* Scene::findLightTransformation()
{
    for (unsigned int i = 0; i < pRootNode->getChildNodeSize(); i++)
    {
        const Node* pNode = pRootNode->getChildNode(i);
        if (pNode->getLightIndex() != "")
        {
            return pNode->getMatrix();
        }
    }
    return new glm::mat4(1.0f);
}

void Scene::insertLightMap(const std::string& key, Light* pLight)
{
    this->mLightMap.insert(
        std::map<std::string, Light*>::value_type(key, pLight));
}

void Scene::setRootBone(Bone* rootBone)
{
    this->pRootBone = rootBone;
}

Bone* Scene::getRootBone()
{
    return this->pRootBone;
}

void Scene::pushAnimation(Animation* pAnimation)
{
    this->mAnimaVec.push_back(pAnimation);
}

Animation* Scene::getAnimation(unsigned int index)
{
    if (this->mAnimaVec.size() <= index)
    {
        return 0;
    }
    return this->mAnimaVec[index];
}

unsigned int Scene::getAnimationCount()
{
    return this->mAnimaVec.size();
}

const Node* Scene::getRootNode() const
{
    return this->pRootNode;
}

void Scene::setRootNode(Node* pNode)
{
    this->pRootNode = pNode;
}

void Scene::pushNode(Node* pNode)
{
    this->mNodeVec.push_back(pNode);
}

Node* Scene::getNode(unsigned int index)
{
    if (this->mNodeVec.size() <= index)
    {
        return 0;
    }
    return this->mNodeVec[index];
}

unsigned int Scene::getNodeSize()
{
    return this->mNodeVec.size();
}

void Scene::setSkin(Skin* skin)
{
    this->pSkin = skin;
}

Skin* Scene::getSkin()
{
    return this->pSkin;
}

bool Scene::hasSkin()
{
    return this->pSkin == 0 ? false : true;
}

char* Scene::getBuffer(unsigned int length)
{
    if (0 == this->pBuffer && 0 != length)
    {
        this->pBuffer = new char[length];
    }
    return this->pBuffer;
}

bool Scene::setBuffer(const std::string& binName, unsigned int length)
{
    int idx = gltf_get_file_index_by_name(pGltfHandle, binName.c_str());
    if (idx >= 0)
    {
        if (pGltfHandle->files[idx]->buffer != NULL)
        {
            memcpy(pBuffer, pGltfHandle->files[idx]->buffer, length);
            return true;
        }
    }
    return false;
}

void Scene::setVertexMax(float x, float y, float z)
{
    mVertexMax.x = std::max(x, mVertexMax.x);
    mVertexMax.y = std::max(y, mVertexMax.y);
    mVertexMax.z = std::max(z, mVertexMax.z);
}

glm::vec3& Scene::getVertexMax()
{
    return mVertexMax;
}

void Scene::setVertexMin(float x, float y, float z)
{
    mVertexMin.x = std::min(x, mVertexMin.x);
    mVertexMin.y = std::min(y, mVertexMin.y);
    mVertexMin.z = std::min(z, mVertexMin.z);
}

glm::vec3& Scene::getVertexMin()
{
    return mVertexMin;
}

Bone* Scene::findBone(const std::string& key)
{
    std::map<std::string, Bone*>::iterator it;
    it = mBoneMap.find(key);
    if (it != mBoneMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertBoneMap(std::string key, Bone* pBone)
{
    this->mBoneMap.insert(
        std::map<std::string, Bone*>::value_type(key, pBone));
}

void Scene::pushTechnique(class Technique* pTechnique)
{
    this->mTechniqueVec.push_back(pTechnique);
}

class Technique* Scene::getTechnique(unsigned int index)
{
    if (this->mTechniqueVec.size() <= index)
    {
        return 0;
    }
    return this->mTechniqueVec[index];
}

unsigned int Scene::getTechSize()
{
    return this->mTechniqueVec.size();
}

bool Scene::loadTexture(const std::string& imagePath)
{
    if (findTexture(imagePath) == NULL)
    {
        Texture* oneTexture = new Texture();

        glTFFile* pglTFFile = getGltfFileByFileName(imagePath);
        oneTexture->loadTexture2DFromBuffer((unsigned char*)
                                            (pglTFFile->buffer),
                                            pglTFFile->imagewidth,
                                            pglTFFile->imageheight);
        oneTexture->setFiltering(TEXTURE_FILTER_MAG_NEAREST,
                                 TEXTURE_FILTER_MIN_NEAREST);
        insertTextureMap(imagePath, oneTexture);
    }
    return true;
}

Texture* Scene::findTexture(const std::string& key)
{
    std::map<std::string, Texture*>::iterator it;
    it = mTexturesMap.find(key);
    if (it != mTexturesMap.end())
    {
        return it->second;
    }
    return NULL;
}

void Scene::insertTextureMap(const std::string& key, Texture* pTexture)
{
    this->mTexturesMap.insert(
        std::map<std::string, Texture*>::value_type(key, pTexture));
}

void Scene::setGltfHandle(glTFHandle* handle)
{
    if (handle != NULL)
    {
        this->pGltfHandle = handle;
    }
}

glTFHandle* Scene::getGltfHandle()
{
    return this->pGltfHandle;
}

glTFFile* Scene::getGltfFileByFileName(const std::string& fileName)
{
    if (pGltfHandle == NULL)
    {
        return NULL;
    }
    for (unsigned int i = 0; i < pGltfHandle->size; i++)
    {
        if (!strcmp(pGltfHandle->files[i]->filename, fileName.c_str()))
        {
            return pGltfHandle->files[i];
        }
    }
    return NULL;
}
}
