/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include <fstream>
#include "LoadScene.h"
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>

#pragma warning(disable : 4996)

Parser::Parser()
{
#if CALC_PARSEING_TIME
    LARGE_INTEGER f;
    QueryPerformanceFrequency(&f);
    dqFreq = (double)f.QuadPart;
#endif
}

Parser::~Parser()
{
}

void Parser::setJsonInfo(const std::string& sbuffer)
{
    this->json_buffer = sbuffer;
    is_json_in_buffer = true;
}

void Parser::setJsonInfo(const std::string& direct, const std::string& fileName)
{
    this->json_directory = direct;
    this->json_file = fileName;
    is_json_in_buffer = false;
}

bool Parser::parseJsonFile()
{
    try
    {
        if (is_json_in_buffer)
        {
#if CALC_PARSEING_TIME
            _LARGE_INTEGER tmp;
            QueryPerformanceCounter(&tmp);
            mParseingStartTime = (double)tmp.QuadPart;
#endif
            std::stringstream ss(this->json_buffer);
            boost::property_tree::json_parser::read_json(ss, ptParse);
#if CALC_PARSEING_TIME
            QueryPerformanceCounter(&tmp);
            mParseingEndTime = (double)tmp.QuadPart;
            ofstream FPSFile = ofstream("Parsing_Time.txt");
            FPSFile << "Parsing Time:" << setprecision (12) <<
                (double)(mParseingEndTime - mParseingStartTime) / dqFreq<< endl;
            FPSFile.close();
#endif
        }
        else
        {
            std::string jsonfile = this->json_directory + this->json_file;
            boost::property_tree::json_parser::read_json(jsonfile.c_str(),
                    ptParse);
        }
    }
    catch (boost::property_tree::ptree_error& e)
    {
        fprintf(stderr, "%s\n", e.what());
        return false;
    }
    catch (boost::exception const&)
    {
        fprintf(stderr, "%s\n", "unknown boost error");
        return false;
    }

    return true;
}

glTFHandle* Parser::getFileNameInJson(const std::string& jsonFile)
{
    glTFHandle* gHandle = (glTFHandle*)malloc(sizeof(glTFHandle));

    boost::property_tree::ptree pImagesTree;
    boost::property_tree::ptree pBuffersTree;
    boost::property_tree::ptree pShadersTree;

    if (isFoundPtree(ptParse, "images"))
    {
        pImagesTree = ptParse.get_child("images");
    }
    if (isFoundPtree(ptParse, "buffers"))
    {
        pBuffersTree = ptParse.get_child("buffers");
    }
    if (isFoundPtree(ptParse, "shaders"))
    {
        pShadersTree = ptParse.get_child("shaders");
    }

    gHandle->size = pImagesTree.size() + pBuffersTree.size() +
                    pShadersTree.size() + 1;
    /* MAYBE: maybe have performance issue.  todo=zhangyang
    if redefine the glTFHandle as following,
    typedef struct
    {
        glTFFile*   files;
        size_t       size;
        glTFViewport viewport;
        void*        renderer;
    } glTFHandle;

    then gHandle->files can be malloc at once; no need malloc with many small chunks.
    */
    unsigned int index = 0;
    gHandle->files = (glTFFile**)malloc(gHandle->size * sizeof(glTFFile*));
    gHandle->files[index] = (glTFFile*)malloc(sizeof(glTFFile));
    gHandle->files[index]->type = GLTF_JSON;
    gHandle->files[index]->filename = (char*)malloc(jsonFile.size() + 1);
    sprintf(gHandle->files[index++]->filename, "%s" , jsonFile.c_str());

    for (boost::property_tree::ptree::iterator it = pImagesTree.begin();
         it != pImagesTree.end(); ++it)
    {
        boost::property_tree::ptree pImageTree = it->second;
        std::string path = pImageTree.get<std::string>("path");
        gHandle->files[index] = (glTFFile*)malloc(sizeof(glTFFile));
        gHandle->files[index]->type = GLTF_IMAGE;
        gHandle->files[index]->filename = (char*)malloc(path.size() + 1);
        sprintf(gHandle->files[index++]->filename, "%s" , path.c_str());
    }

    for (boost::property_tree::ptree::iterator it = pBuffersTree.begin();
         it != pBuffersTree.end(); ++it)
    {
        boost::property_tree::ptree pBufferTree = it->second;
        std::string path = pBufferTree.get<std::string>("path");
        gHandle->files[index] = (glTFFile*)malloc(sizeof(glTFFile));
        gHandle->files[index]->type = GLTF_BINARY;
        gHandle->files[index]->filename = (char*)malloc(path.size() + 1);
        sprintf(gHandle->files[index++]->filename, "%s" , path.c_str());
    }

    for (boost::property_tree::ptree::iterator it = pShadersTree.begin();
         it != pShadersTree.end(); ++it)
    {
        boost::property_tree::ptree pShaderTree = it->second;
        std::string path = pShaderTree.get<std::string>("path");
        gHandle->files[index] = (glTFFile*)malloc(sizeof(glTFFile));
        gHandle->files[index]->type = GLTF_GLSL;
        gHandle->files[index]->filename = (char*)malloc(path.size() + 1);
        sprintf(gHandle->files[index++]->filename, "%s" , path.c_str());
    }

    return gHandle;
}

bool Parser::releaseFileName()
{
    glTFHandle* gHandle = pScene->getGltfHandle();
    for (int i = (int)gHandle->size - 1 ; i >= 0 ; i--)
    {
        if (gHandle->files[i]->filename != NULL)
        {
            free(gHandle->files[i]->filename);
            gHandle->files[i]->filename = NULL;
        }
        if (gHandle->files[i] != NULL)
        {
            free(gHandle->files[i]);
            gHandle->files[i] = NULL;
        }
    }

    if (gHandle->files != NULL)
    {
        free(gHandle->files);
        gHandle->files = NULL;
    }
    delete gHandle;
    return true;
}
bool Parser::parseScene(Scene* pscene)
{
    if (0 == pscene)
    {
        return false;
    }
    this->pScene = pscene;

    if (!this->readBuffers())
    {
        return false;
    }
    if (!this->parseNodes())
    {
        return false;
    }
    if (!this->parseMeshs())
    {
        return false;
    }
    if (!this->parseMaterials())
    {
        return false;
    }
    if (ptParse.find("cameras") != ptParse.not_found())
    {
        if (!this->parseCameras())
        {
            return false;
        }
    }
    if (ptParse.find("lights") != ptParse.not_found())
    {
        if (!this->parseLights())
        {
            return false;
        }
    }
    if (!this->parseAttributes())
    {
        return false;
    }
    if (!this->parseTechniques())
    {
        return false;
    }
    if (!this->parseAnim())
    {
        return false;
    }

    return true;
}

bool Parser::parseNodes()
{
    glm::mat4 matrix(1.0f);
    Node* pRootNode = new Node();
    Bone* pRootBone = new Bone();
    if(!isFoundPtree(ptParse,"scene"))
    {
        return false;
    }
    const std::string& sceneId = ptParse.get<std::string>("scene");
    std::string pathNode = "scenes." + sceneId + ".nodes";
    const boost::property_tree::ptree& pNodeArray =
        ptParse.get_child(pathNode.c_str());
    for (boost::property_tree::ptree::const_iterator it = pNodeArray.begin();
        it != pNodeArray.end(); ++it)
    {
        this->parseNode(it->second.data(), pRootNode, matrix);
        this->parseBone(it->second.data(), pRootBone);
    }
    this->pScene->setRootNode(pRootNode);
    this->pScene->setRootBone(pRootBone);
    return true;
}

void Parser::parseBone(const std::string& id, Bone* pParentBone)
{
    std::string nodeId = "nodes." + id;
    const boost::property_tree::ptree& pBoneAtrr = ptParse.get_child(nodeId.c_str());
    if (pBoneAtrr.find("jointId") == pBoneAtrr.not_found())
    {
        return;
    }
    const std::string& jointId = pBoneAtrr.get<std::string>("jointId");
    Bone* pBone = new Bone();
    pBone->setJointId(jointId);
    boost::property_tree::ptree::const_iterator it;
    boost::property_tree::ptree::const_iterator itEnd;
    boost::property_tree::ptree::const_assoc_iterator itAttr=pBoneAtrr.find("name");
    if (itAttr != pBoneAtrr.not_found())
    {
        pBone->setNodeName((itAttr->second).get_value<std::string>("name"));
    }
    itAttr = pBoneAtrr.find("matrix");
    if (itAttr != pBoneAtrr.not_found())
    {
        float matrix[16];
        unsigned int i = 0;
        const boost::property_tree::ptree& pMatrix = itAttr->second;
        itEnd = pMatrix.end();
        for (it = pMatrix.begin(); it !=itEnd; ++it)
        {
            matrix[i++] = (it->second).get_value<float>();
        }
        pBone->setLocalMatrix(matrix);
    }
    itAttr = pBoneAtrr.find("rotation");
    if (itAttr != pBoneAtrr.not_found())
    {
      const boost::property_tree::ptree& pRotate = itAttr->second;
      float rotate[4];
      unsigned int i = 0;
      itEnd = pRotate.end();
      for (it = pRotate.begin();it !=itEnd; ++it)
      {
        rotate[i++] = (it->second).get_value<float>();
      }
      pBone->setRotate(rotate);
      pBone->setJointFlag(true);
    }
    itAttr = pBoneAtrr.find("scale");
    if (itAttr != pBoneAtrr.not_found())
    {
      const boost::property_tree::ptree& pScale = itAttr->second;
      float scale[3];
      unsigned int i = 0;
      itEnd = pScale.end();
      for (it = pScale.begin();it !=itEnd; ++it)
      {
        scale[i++] = (it->second).get_value<float>();
      }
      pBone->setScale(scale);
      pBone->setJointFlag(true);
    }
    itAttr = pBoneAtrr.find("translation");
    if ( itAttr!= pBoneAtrr.not_found())
    {
        const boost::property_tree::ptree& pTrans = itAttr->second;
        float trans[3];
        unsigned int i = 0;
        itEnd =  pTrans.end();
        for (it = pTrans.begin(); it != itEnd; ++it)
        {
            trans[i++] = (it->second).get_value<float>();
        }
        pBone->setTranslate(trans);
        pBone->setJointFlag(true);
    }
    itAttr = pBoneAtrr.find("children");
    if (itAttr != pBoneAtrr.not_found()
        && 0 !=itAttr->second.size())
    {
        const boost::property_tree::ptree& pBoneChildren=itAttr->second;
        itEnd = pBoneChildren.end();
        for (it = pBoneChildren.begin(); it != itEnd; ++it)
        {
            parseBone(it->second.data(), pBone);
        }
    }
    pParentBone->pushChildBone(pBone);
    pBone->setParentBone(pParentBone);
    pScene->insertBoneMap(jointId, pBone);
    return;
}

void Parser::parseNode(const std::string& id, Node* pParentNode,
                       const glm::mat4& parentMatrix)
{
    std::string nodeId = "nodes." + id;
    const boost::property_tree::ptree& pNodeAtrr = ptParse.get_child(nodeId.c_str());
    boost::property_tree::ptree::const_assoc_iterator itAttr=pNodeAtrr.find("jointId");
    if (itAttr!= pNodeAtrr.not_found())
    {
        return;
    }
    
    Node* pNode = new Node();
    itAttr = pNodeAtrr.find("name");
    if (itAttr!= pNodeAtrr.not_found())
    {
        pNode->setNodeName(itAttr->second.get_value<std::string>("name"));
    }

    boost::property_tree::ptree::const_iterator it;
    boost::property_tree::ptree::const_iterator itEnd;
    itAttr = pNodeAtrr.find("meshes");
    if ( itAttr!= pNodeAtrr.not_found())
    {
        pNode->setNodeType(NodeType_Mesh);
        const boost::property_tree::ptree& pMeshes = itAttr->second;
        itEnd = pMeshes.end();
        for (it = pMeshes.begin(); it != itEnd; ++it)
        {
            pNode->pushMeshIndex(it->second.data());
        }
    }
    else if ( (itAttr = pNodeAtrr.find("light")) != pNodeAtrr.not_found())
    {
        pNode->setLightIndex(itAttr->second.get_value<std::string>());
        pNode->setNodeType(NodeType_Light);
        pNode->setLightFlag();
    }
    else if ((itAttr = pNodeAtrr.find("camera")) != pNodeAtrr.not_found())
    {
        pNode->setCameraIndex(itAttr->second.get_value<std::string>());
        pNode->setNodeType(NodeType_Camera);
    }
    else if ((itAttr = pNodeAtrr.find("instanceSkin")) != pNodeAtrr.not_found())
    {
        const boost::property_tree::ptree& pInstSkin = itAttr->second;
        const boost::property_tree::ptree& pSources = pInstSkin.get_child("sources");
        itEnd =  pSources.end();
        for (it = pSources.begin(); it != itEnd; ++it)
        {
            pNode->pushMeshIndex(it->second.data());
        }

        const std::string &skinId = pInstSkin.get<std::string>("skin");

        Skin* pSkin = this->parseSkins(skinId);
        this->pScene->setSkin(pSkin);
    }

    glm::mat4* pmatrix = (glm::mat4*) new glm::mat4;
    *pmatrix = glm::mat4(1.0f);
    itAttr = pNodeAtrr.find("matrix");
    if ( itAttr != pNodeAtrr.not_found())
    {
        const boost::property_tree::ptree& pMatrix = itAttr->second;
        float matrix[16];
        unsigned int i = 0;
        itEnd = pMatrix.end();
        for (it = pMatrix.begin();it != itEnd; ++it)
        {
            matrix[i++] = (it->second).get_value<float>();
        }
        memcpy(pmatrix, matrix, sizeof(glm::mat4));
        pNode->setCurrentNodeMatrix(*pmatrix);
        *pmatrix = parentMatrix * (*pmatrix);
        pNode->setMatrix(pmatrix);
    }

    pParentNode->pushChildNode(pNode);
    pNode->setParentNode(pParentNode);
    pNode->setParentMatrixes();
    this->pScene->pushNode(pNode);

    itAttr = pNodeAtrr.find("children");
   
    if ( itAttr!= pNodeAtrr.not_found()
        && 0 != pNodeAtrr.get_child("children").size())
    {
        pNode->setNodeType(NodeType_Node);
        const boost::property_tree::ptree& pNodeChildren = itAttr->second;
        itEnd = pNodeChildren.end();
        for (it = pNodeChildren.begin(); it !=itEnd; ++it)
        {
            parseNode(it->second.data(), pNode, *pmatrix);
        }
    }
    return;
}

Skin* Parser::parseSkins(const std::string& skinId)
{
    const boost::property_tree::ptree& pSkinsTree = ptParse.get_child("skins");
    if (pSkinsTree.find(skinId.c_str()) == pSkinsTree.not_found())
    {
        return 0;
    }
    Skin* pSkin = new Skin();

    std::string shapeMatrixId = skinId + "." + "bindShapeMatrix";
    const boost::property_tree::ptree& pShapeTree =
        pSkinsTree.get_child(shapeMatrixId.c_str());
    float buffer[16];
    unsigned int i = 0;
    boost::property_tree::ptree::const_iterator it;
    boost::property_tree::ptree::const_iterator itEnd = pShapeTree.end();
    for (it = pShapeTree.begin(); it !=itEnd; ++it)
    {
        buffer[i++] = (it->second).get_value<float>();
    }

    glm::mat4 shape(1.0);
    memcpy(&shape, buffer, sizeof(glm::mat4));

    std::string bindMatrixId = skinId + "." + "inverseBindMatrices";
    const boost::property_tree::ptree& pBind =
        pSkinsTree.get_child(bindMatrixId.c_str());
    std::string bufferView =
        "bufferViews." + pBind.get<std::string>("bufferView");
    const boost::property_tree::ptree& pBufferView =
        ptParse.get_child(bufferView.c_str());
    char* pbuffer = this->pScene->getBuffer(0)
                    + pBufferView.get<unsigned int>("byteOffset")
                    + pBind.get<unsigned int>("byteOffset");
    unsigned int count = pBind.get<unsigned int>("count");
    glm::mat4* pBindArray = new glm::mat4[count];
    for (unsigned int i = 0; i < count; i++)
    {
        memcpy(&pBindArray[i], pbuffer + i * 64, sizeof(glm::mat4));
        pBindArray[i] = pBindArray[i] * shape;
    }
    pSkin->setBindMatrix(pBindArray);
    pSkin->setBindMatrixCount(count);

    std::string jointId = skinId + "." + "joints";
    const boost::property_tree::ptree& pJoint = pSkinsTree.get_child(jointId.c_str());
    itEnd = pJoint.end();
    for (it = pJoint.begin(); it != itEnd; ++it)
    {
        pSkin->pushBoneId((it->second).get_value<std::string>());
    }
    return pSkin;
}

bool Parser::parseMeshs()
{
    const boost::property_tree::ptree& pMeshsTree = ptParse.get_child("meshes");
    boost::property_tree::ptree::const_iterator it = pMeshsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd =  pMeshsTree.end();
    for (;it != itEnd; ++it)
    {
        Mesh* pMesh = new Mesh();
        const boost::property_tree::ptree& pMeshTree = it->second;
        pMesh->setMeshName(pMeshTree.get<std::string>("name"));
        parsePrimitive(pMeshTree.get_child("primitives"), pMesh);

        this->pScene->insertMeshMap(it->first, pMesh);
    }
    return true;
}

bool Parser::parsePrimitive(const boost::property_tree::ptree& pTree, Mesh* pMesh)
{
    boost::property_tree::ptree::const_iterator it = pTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pTree.end();
    for (;it !=itEnd; ++it)
    {
        Primitives* pPrim = new Primitives();

        const boost::property_tree::ptree& pPrimTree = it->second;
        const boost::property_tree::ptree& pPrimValueTree =
          pPrimTree.get_child("attributes");
        boost::property_tree::ptree::const_assoc_iterator itPVT = 
          pPrimValueTree.find("NORMAL");
        if ( itPVT != pPrimValueTree.not_found())
        {
            pPrim->insertAttribute("NORMAL",
                                   itPVT->second.get_value<std::string>());
        }
        itPVT = pPrimValueTree.find("POSITION");
        if (itPVT != pPrimValueTree.not_found())
        {
            pPrim->insertAttribute("POSITION",
                                   itPVT->second.get_value<std::string>());
        }
        itPVT = pPrimValueTree.find("TEXCOORD_0");
        if ( itPVT != pPrimValueTree.not_found())
        {
            pPrim->insertAttribute("TEXCOORD_0",
                                   itPVT->second.get_value<std::string>());
        }
        itPVT = pPrimValueTree.find("JOINT");
        if ( itPVT != pPrimValueTree.not_found())
        {
            pPrim->insertAttribute("JOINT",
                                   itPVT->second.get_value<std::string>());
        }
        itPVT = pPrimValueTree.find("WEIGHT");
        if ( itPVT != pPrimValueTree.not_found())
        {
            pPrim->insertAttribute("WEIGHT",
                                   itPVT->second.get_value<std::string>());
        }
        pPrim->setIndicesIndex(pPrimTree.get<std::string>("indices"));
        pPrim->setMaterialIndex(pPrimTree.get<std::string>("material"));
        pMesh->setPrimitiveVec(pPrim);
    }
    return true;
}

bool Parser::parseMaterials()
{
    const boost::property_tree::ptree& pMaterialsTree =
        ptParse.get_child("materials");
    boost::property_tree::ptree::const_iterator it = pMaterialsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pMaterialsTree.end();
    for (;it != itEnd; ++it)
    {
        Material* pMaterial = new Material();

        const boost::property_tree::ptree& pMaterialTree = it->second;
        const std::string& techniqueId =
            pMaterialTree.get<std::string>("instanceTechnique.technique");
        parseMaterialProper(
            pMaterialTree.get_child("instanceTechnique.values"),
            pMaterial, techniqueId);

        pMaterial->setTechniqueId(techniqueId);
        this->pScene->insertMaterialMap(it->first, pMaterial);
    }
    return true;
}

bool Parser::parseMaterialProper(const boost::property_tree::ptree& pTree,
                                 Material* pMaterial,
                                 const std::string& techniqueId)
{
    std::string properName;
    boost::property_tree::ptree::const_iterator it = pTree.begin();
    boost::property_tree::ptree::const_iterator itEnd =  pTree.end();
    for (;it != itEnd; ++it)
    {
        MaterialProperty* pMaterialProper = new MaterialProperty();

        std::string typePath = "techniques." + techniqueId +
                               ".parameters." + it->first + ".type";
        unsigned int type = ptParse.get<unsigned int>(typePath.c_str());
        properName = "u_" + it->first;
        pMaterialProper->setPropertyName(properName);
        switch (type)
        {
            case DataType_SAMPLER_2D:
                {
                    const std::string& textureId = it->second.data();
                    std::string textureSource =
                        "textures." + textureId + ".source";
                    const std::string& imageId =
                        ptParse.get<std::string>(textureSource.c_str());
                    std::string imagePath = "images." + imageId + ".path";
                    pMaterialProper->setImagePath(
                        ptParse.get<std::string>(imagePath.c_str()));
                    pMaterialProper->setDataLength(0);
                    pMaterialProper->setDataType(DataType_SAMPLER_2D);
                    pScene->loadTexture(pMaterialProper->getImagePath());
                }
                break;
            case DataType_FLOAT:
                {
                    pMaterialProper->setDataLength(sizeof(float));
                    // todo=qulei design cause maybe memory leak;
                    float data = 0;
                    data = (it->second).get_value<float>();
                    pMaterialProper->setPropertyData((char*)&data, sizeof(float));
                    pMaterialProper->setDataType(DataType_FLOAT);
                }
                break;
            case DataType_FLOAT_VEC4:
                {
                    pMaterialProper->setDataLength(sizeof(float) * 4);
                    const boost::property_tree::ptree& pValue = it->second;
                    float buffer[4] = {0};
                    boost::property_tree::ptree::const_iterator iter = pValue.begin();
                    boost::property_tree::ptree::const_iterator iterEnd = pValue.end();
                    for (unsigned int i = 0;(i < 4) && (iter!=iterEnd) ; ++iter, ++i)
                    {
                        buffer[i] = (iter->second).get_value<float>();
                    }
                    pMaterialProper->setPropertyData((char*)buffer, 4*sizeof(float));
                    pMaterialProper->setDataType(DataType_FLOAT_VEC4);
                }
                break;
            default:
                break;
        }
        pMaterial->pushMaterialProper(pMaterialProper);
    }

    return true;
}

bool Parser::parseCameras()
{
    const boost::property_tree::ptree& pCamerasTree = ptParse.get_child("cameras");
    boost::property_tree::ptree::const_iterator it = pCamerasTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pCamerasTree.end();
    for (;it != itEnd; ++it)
    {
        ParseCamera* pCamera = new ParseCamera();

        const boost::property_tree::ptree& pCameraTree = it->second;
        const std::string& type = pCameraTree.get<std::string>("type");
        const boost::property_tree::ptree& pCameraAttrTree =
            pCameraTree.get_child(type.c_str());
        boost::property_tree::ptree::const_assoc_iterator itAttr= pCameraAttrTree.find("aspect_ratio");
        if (itAttr != pCameraAttrTree.not_found())
        {
            pCamera->setXFov(itAttr->second.get_value<float>());
        }
        else if ( (itAttr = pCameraAttrTree.find("xfov")) !=
            pCameraAttrTree.not_found())
        {
            pCamera->setYFov(itAttr->second.get_value<float>());
        }
        else if ( (itAttr = pCameraAttrTree.find("yfov")) !=
            pCameraAttrTree.not_found())
        {
            pCamera->setYFov(itAttr->second.get_value<float>());
        }
        else if ( (itAttr = pCameraAttrTree.find("zfar")) !=
            pCameraAttrTree.not_found())
        {
            pCamera->setYFov(itAttr->second.get_value<float>());
        }
        else if ( (itAttr = pCameraAttrTree.find("znear")) !=
            pCameraAttrTree.not_found())
        {
            pCamera->setYFov(itAttr->second.get_value<float>());
        }
        this->pScene->insertCameraMap(it->first, pCamera);
    }
    return true;
}

Light* GetParseLight(const boost::property_tree::ptree&  pParseLight
                     , LightSourceType_E NodeType)
{
    unsigned int i;
    float fLight[4];
    memset(fLight, 0, sizeof(float) * 4);
    Light* pLight = new Light();
    pLight->setType(NodeType);
    boost::property_tree::ptree::const_iterator it;
    boost::property_tree::ptree::const_iterator itEnd;
    boost::property_tree::ptree::const_assoc_iterator itL =pParseLight.find("color");
    if ( itL!= pParseLight.not_found())
    {
        i = 0;
        const boost::property_tree::ptree& color = itL->second;
        it = color.begin();
        itEnd = color.end();
        for (;it != itEnd; it++, i++)
        {
            fLight[i] = (it->second).get_value<float>();
        }
        pLight->setColorAmbient((GLfloat*)&fLight[0]);
        pLight->setColorDiffuse((GLfloat*)&fLight[1]);
        pLight->setColorSpecular((GLfloat*)&fLight[2]);
    }
    itL = pParseLight.find("position");
    if ( itL!= pParseLight.not_found())
    {
        i = 0;
        const boost::property_tree::ptree& pos =itL->second;
        it = pos.begin();
        itEnd = pos.end();
        for (;it != itEnd; it++, i++)
        {
            fLight[i] = (it->second).get_value<float>();
        }
        if (i >= 3)
        {
            pLight->setPositions((GLfloat*)fLight);
        }
        else
        {
            glm::vec3 LightPos(fLight[0], fLight[1], fLight[2]);
            pLight->setPosition(LightPos);
            pLight->setPosFlag();
        }
    }
    return pLight;
}

bool Parser::parseLights()
{
    const boost::property_tree::ptree& pLightsTree = ptParse.get_child("lights");
    boost::property_tree::ptree::const_iterator it = pLightsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd =  pLightsTree.end();
    for ( ;it !=itEnd; ++it)
    {
        Light* pLight = new Light();
        const boost::property_tree::ptree& pLightTree = it->second;
        boost::property_tree::ptree::const_assoc_iterator itL = pLightTree.find("point");
        if ( itL != pLightTree.not_found())
        {
            pLight = GetParseLight(itL->second, LightSource_POINT);
        }
        else if ( (itL =pLightTree.find("directional")) != pLightTree.not_found())
        {
            pLight = GetParseLight(itL->second, LightSource_DIRECTIONAL);
        }
    }
    return true;
}

bool Parser::parseAttributes()
{
    const boost::property_tree::ptree& pAttrsTree = ptParse.get_child("accessors");
    boost::property_tree::ptree::const_iterator it = pAttrsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pAttrsTree.end();
    for (;it != itEnd; ++it)
    {
        Attribute* pAttribute = new Attribute();
        const boost::property_tree::ptree& pAttrTree = it->second;
        unsigned int type = pAttrTree.get<unsigned int>("type");
        pAttribute->setDataType((enum DataType)type);

        unsigned int byteStride;
        switch (type)
        {
            case DataType_UNSIGNED_SHORT:
                byteStride = 2;
                break;
            case DataType_FLOAT:
                byteStride = 4;
                break;
            case DataType_FLOAT_VEC2:
                byteStride = 8;
                break;
            case DataType_FLOAT_VEC3:
                byteStride = 12;
                break;
            case DataType_FLOAT_VEC4:
                byteStride = 16;
                break;
            default:
                byteStride = 4;
                break;
        }
        pAttribute->setByteStride(byteStride);

        unsigned int count = pAttrTree.get<unsigned int>("count");
        pAttribute->setDataCount(count);

        std::string bufferView = "bufferViews." +
                                 pAttrTree.get<std::string>("bufferView");
        const boost::property_tree::ptree& pBufferView =
            ptParse.get_child(bufferView.c_str());
        char* pbuffer = this->pScene->getBuffer(0)
                        + pBufferView.get<unsigned int>("byteOffset")
                        + pAttrTree.get<unsigned int>("byteOffset");
        pAttribute->setAttributeData(pbuffer, count * byteStride);

        
        if (DataType_FLOAT_VEC3 == type)
        {
           boost::property_tree::ptree::const_assoc_iterator itAttr = pAttrTree.find("max");
           if(itAttr!= pAttrTree.not_found())
           {
              float max[3] = {0.f, 0.f, 0.f};
              const boost::property_tree::ptree& pMaxTree = itAttr->second;
              boost::property_tree::ptree::const_iterator iter = pMaxTree.begin();
              boost::property_tree::ptree::const_iterator iterEnd = pMaxTree.end();
              for (unsigned int i = 0; (iter != iterEnd) && (i<3); ++iter, ++i)
              {
                  max[i] = (iter->second).get_value<float>();
              }
              this->pScene->setVertexMax(max[0], max[1], max[2]);
          }
          itAttr = pAttrTree.find("min");
          if ( itAttr!= pAttrTree.not_found())
          {
              float min[3] = {0.f, 0.f, 0.f};
              const boost::property_tree::ptree& pMinTree = itAttr->second;
              boost::property_tree::ptree::const_iterator iter = pMinTree.begin();
              boost::property_tree::ptree::const_iterator iterEnd = pMinTree.end();
              for (unsigned int i = 0; (iter != iterEnd) && (i<3); ++iter, ++i)
              {
                  min[i++] = (iter->second).get_value<float>();
              }
              this->pScene->setVertexMin(min[0], min[1], min[2]);
          }
        }
        this->pScene->insertAttributeMap(it->first, pAttribute);
    }
    return true;
}

bool Parser::parseTechniques()
{
    const boost::property_tree::ptree& pTechniquesTree =
        ptParse.get_child("techniques");
    boost::property_tree::ptree::const_iterator it = pTechniquesTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pTechniquesTree.end();
    for (;it != itEnd; ++it)
    {
        Technique* pTechnique = new Technique();

        const boost::property_tree::ptree& pTechniqueTree = it->second;
        const std::string& passId = pTechniqueTree.get<std::string>("pass");
        std::string passPath = "passes." + passId + ".instanceProgram";
        const boost::property_tree::ptree& pInstanceTree =
            pTechniqueTree.get_child(passPath.c_str());
        const boost::property_tree::ptree& pParameterTree =
            pTechniqueTree.get_child("parameters");
        parseTechniqueProgram(pInstanceTree, pParameterTree, pTechnique);

        std::string statePath = "passes." + passId + ".states";
        boost::property_tree::ptree pStateTree =
            pTechniqueTree.get_child(statePath.c_str());
        parseTechniqueState(pStateTree, pTechnique);
        pTechnique->setTechId(it->first);
        this->pScene->pushTechnique(pTechnique);
    }
    return true;
}

bool Parser::parseTechniqueState(boost::property_tree::ptree& pInstanceTree,
                                 Technique* pTechnique)
{
    for (boost::property_tree::ptree::iterator it = pInstanceTree.begin();
         it != pInstanceTree.end(); ++it)
    {
        if ("blendFunc" == it->first)
        {
            boost::property_tree::ptree dfactorNode =
                pInstanceTree.get_child("blendFunc.dfactor");
            boost::property_tree::ptree sfactorNode =
                pInstanceTree.get_child("blendFunc.sfactor");
            pTechnique->getTechState()->blendFuncDfactor =
                dfactorNode.get_value<unsigned int>();
            pTechnique->getTechState()->blendFuncSfactor =
                sfactorNode.get_value<unsigned int>();
            continue;
        }
        if ("blendEnable" == it->first)
        {
            pTechnique->getTechState()->blendEnable =
                it->second.get_value<unsigned int>();
            continue;
        }
        if ("blendEquation" == it->first)
        {
            pTechnique->getTechState()->blendEquation =
                it->second.get_value<unsigned int>();
            continue;
        }
        if ("cullFaceEnable" == it->first)
        {
            pTechnique->getTechState()->cullFaceEnable =
                it->second.get_value<unsigned int>();
            continue;
        }
        if ("depthMask" == it->first)
        {
            pTechnique->getTechState()->depthMask =
                it->second.get_value<unsigned int>();
            continue;
        }
        if ("depthTestEnable" == it->first)
        {
            pTechnique->getTechState()->depthTestEnable =
                it->second.get_value<unsigned int>();
            continue;
        }
    }
    return true;
}

int Parser::parseTechniqueProgram(const boost::property_tree::ptree& pInstanceTree,
                                   const boost::property_tree::ptree& pParameterTree,
                                   Technique* pTechnique)
{
    boost::property_tree::ptree pAttrsTree =
        pInstanceTree.get_child("attributes");
    boost::property_tree::ptree::const_iterator it = pAttrsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pAttrsTree.end();
    for (; it != itEnd; ++it)
    {
        const std::string& parameterId = it->second.data();
        const boost::property_tree::ptree& pParaTree =
            pParameterTree.get_child(parameterId.c_str());
        boost::property_tree::ptree::const_assoc_iterator itPT = 
          pParaTree.find("semantic");
        if ( itPT != pParaTree.not_found())
        {
            TechAttribute* pTechAttribute = new TechAttribute();
            pTechAttribute->setAttributeIndex(
               itPT->second.get_value<std::string>());
            pTechAttribute->setAttributeName(it->first);
            pTechnique->insertTechAttribute(it->first, pTechAttribute);
        }
    }
    const boost::property_tree::ptree& pUniformsTree =
        pInstanceTree.get_child("uniforms");
    it = pUniformsTree.begin();
    itEnd = pUniformsTree.end();
    for (; it !=itEnd; ++it)
    {
        TechUniform* pTechUniform = new TechUniform();
        pTechUniform->setUniformIndex(it->second.data());
        pTechUniform->setUniformName(it->first);
        pTechnique->pushTechUniform(pTechUniform);
    }
    std::string programId = "programs." +
                            pInstanceTree.get<std::string>("program");
    const boost::property_tree::ptree& pProgramTree =
        ptParse.get_child(programId.c_str());
    std::string verId = "shaders." +
                        pProgramTree.get<std::string>("vertexShader") +
                        ".path";
    std::string fraId = "shaders." +
                        pProgramTree.get<std::string>("fragmentShader") +
                        ".path";
    pTechnique->setVertexShader(ptParse.get<std::string>(verId.c_str()));
    pTechnique->setFragmentShader(ptParse.get<std::string>(fraId.c_str()));
    return pTechnique->initTechnique(pScene->getGltfHandle()); 
}

bool Parser::parseAnim()
{
    const boost::property_tree::ptree pAnimsTree = ptParse.get_child("animations");
    boost::property_tree::ptree::const_iterator it = pAnimsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pAnimsTree.end();
    for (;it != itEnd; ++it)
    {
        Animation* pAnimation = new Animation();
        boost::property_tree::ptree pAnimTree = it->second;
        parsePara(pAnimTree, pAnimation);
        pAnimation->setBoneId(parseChannel(pAnimTree));
        this->pScene->pushAnimation(pAnimation);
    }
    return true;
}

std::string Parser::parseChannel(const boost::property_tree::ptree& pTree)
{
    std::string id;
    const boost::property_tree::ptree& pNodes = ptParse.get_child("nodes");
    const boost::property_tree::ptree& pChannsTree = pTree.get_child("channels");
    boost::property_tree::ptree::const_iterator it = pChannsTree.begin();
    boost::property_tree::ptree::const_iterator itEnd = pChannsTree.end();
    for (;it != itEnd; ++it)
    {
        const boost::property_tree::ptree pChannTree = it->second;
        std::string idPath = pChannTree.get<std::string>("target.id") +
                             ".jointId";
        id = pNodes.get<std::string>(idPath.c_str());
        break;
    }
    return id;
}

int Parser::parsePara(const boost::property_tree::ptree& pTree,
                      Animation* pAnimation)
{
    unsigned int count = pTree.get<unsigned int>("count");
    const std::string& timeIndex = pTree.get<std::string>("parameters.TIME");
    const Attribute* pTimeAttr = this->pScene->findAttribute(timeIndex);
    if (0 == pTimeAttr)
    {
        return -1;
    }
    const float* pTimeBuf = (const float*)pTimeAttr->getAttributeData();

    pAnimation->setDuration(pTimeBuf[count-1]-pTimeBuf[0]);

    const boost::property_tree::ptree pParaTree = pTree.get_child("parameters");
    boost::property_tree::ptree::const_assoc_iterator itAttr = pParaTree.find("rotation");
    if (itAttr != pParaTree.not_found())
    {
        const std::string& rotaIndex = (itAttr->second).get_value<std::string>("rotation");
        const Attribute* pRotaAttr = this->pScene->findAttribute(rotaIndex);
        if (0 != pRotaAttr)
        {
            AnimChannel* pChannel = new AnimChannel();
            const char* pRotaBuf  = pRotaAttr->getAttributeData();
            unsigned int sizeVec4 = sizeof(glm::vec4);
            for (unsigned int index = 0; index < count; ++index)
            {
                glm::vec4* pValue = (glm::vec4*) new glm::vec4;
                memcpy(pValue, pRotaBuf+index*sizeVec4, sizeVec4);
                pChannel->pushTimeValue4(pTimeBuf[index], pValue);
            }
            pAnimation->setRotateChl(pChannel);
        }
    }
    itAttr = pParaTree.find("scale");
    if (itAttr != pParaTree.not_found())
    {
        const std::string& scaleIndex = (itAttr->second).get_value<std::string>("scale");
        const Attribute* pScaleAttr = this->pScene->findAttribute(scaleIndex);
        if (0 != pScaleAttr)
        {
            AnimChannel* pChannel = new AnimChannel();
            const char* pScaleBuf  = pScaleAttr->getAttributeData();
            unsigned int sizeVec3 = sizeof(glm::vec3);
            for (unsigned int index = 0; index < count; ++index)
            {
                glm::vec3* pValue = (glm::vec3*) new glm::vec3;
                memcpy(pValue, pScaleBuf+index*sizeVec3, sizeVec3);
                pChannel->pushTimeValue3(pTimeBuf[index], pValue);
            }
            pAnimation->setScaleChl(pChannel);
        }
    }
    itAttr = pParaTree.find("translation");
    if (itAttr != pParaTree.not_found())
    {
        const std::string& transIndex = (itAttr->second).get_value<std::string>("translation");
        const Attribute* pTransAttr = this->pScene->findAttribute(transIndex);
        if (0 != pTransAttr)
        {
            AnimChannel* pChannel = new AnimChannel();
            const char* pTransBuf  = pTransAttr->getAttributeData();
            unsigned int sizeVec3 = sizeof(glm::vec3);
            for (unsigned int index = 0; index < count; ++index)
            {
                glm::vec3* pValue = (glm::vec3*) new glm::vec3;
                memcpy(pValue, pTransBuf+index*sizeVec3, sizeVec3);
                pChannel->pushTimeValue3(pTimeBuf[index], pValue);
            }
            pAnimation->setTransChl(pChannel);
        }
    }

    return 0;
}

bool Parser::readBuffers()
{
    unsigned int length = 0;
    std::string binName;
    const  boost::property_tree::ptree& pBuffers = ptParse.get_child("buffers");
    boost::property_tree::ptree::const_iterator it = pBuffers.begin();
    boost::property_tree::ptree::const_iterator itEnd = pBuffers.end();
    for (; it != itEnd; ++it)
    {
        const boost::property_tree::ptree& pbuffer = it->second;
        binName = pbuffer.get<std::string>("path");
        length = pbuffer.get<unsigned int>("byteLength");
    }
    char* pbuffer = this->pScene->getBuffer(length);
    if (!is_json_in_buffer)
    {
        std::fstream file;
        std::string binfile = this->json_directory + binName;
        file.open(binfile.c_str(), std::ios::binary | std::ios::in);
        file.read(pbuffer, length);
        file.close();
        return true;
    }
    else
    {
        return this->pScene->setBuffer(binName, length);
    }
    return false;
}
