/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "RenderScene.h"
#include <glm/gtx/transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>
#include <cerrno>

RenderPrimitive::RenderPrimitive()
    : mVerterCount(0), mIndicesCount(0)
    , pMatrix(0), pMaterial(0)
{
}

RenderPrimitive::~RenderPrimitive()
{
    glDeleteBuffers(1, &mVertexBuffer);
    glDeleteBuffers(1, &mNormalBuffer);
    glDeleteBuffers(1, &mTexCoordBuffer);
    glDeleteBuffers(1, &mIndicesBuffer);
}

Material* RenderPrimitive::getMaterial()
{
    return this->pMaterial;
}

void RenderPrimitive::setMaterial(Material* pMaterial)
{
    this->pMaterial = pMaterial;
}

unsigned int RenderPrimitive::getVerterCount()
{
    return this->mVerterCount;
}

void RenderPrimitive::setVerterCount(unsigned int count)
{
    this->mVerterCount = count;
}

unsigned int RenderPrimitive::getIndicesCount()
{
    return this->mIndicesCount;
}

void RenderPrimitive::setIndicesCount(unsigned int count)
{
    this->mIndicesCount = count;
}

DataType_E RenderPrimitive::getIndicesDataType()
{
    return this->mIndicesDataType;
}

void RenderPrimitive::setIndicesDataType(DataType_E type)
{
    this->mIndicesDataType = type;
}

unsigned int RenderPrimitive::getVertexBuffer()
{
    return this->mVertexBuffer;
}

void RenderPrimitive::setVertexBuffer(unsigned int bufferId)
{
    this->mVertexBuffer = bufferId;
}

unsigned int RenderPrimitive::getNormalBuffer()
{
    return this->mNormalBuffer;
}

void RenderPrimitive::setNormalBuffer(unsigned int bufferId)
{
    this->mNormalBuffer = bufferId;
}

unsigned int RenderPrimitive::getTexCoordBuffer()
{
    return this->mTexCoordBuffer;
}

void RenderPrimitive::setTexCoordBuffer(unsigned int bufferId)
{
    this->mTexCoordBuffer = bufferId;
}

unsigned int RenderPrimitive::getJointBuffer()
{
    return this->mJointBuffer;
}

void RenderPrimitive::setJointBuffer(unsigned int bufferId)
{
    this->mJointBuffer = bufferId;
}

unsigned int RenderPrimitive::getWeightBuffer()
{
    return this->mWeightBuffer;
}

void RenderPrimitive::setWeightBuffer(unsigned int bufferId)
{
    this->mWeightBuffer = bufferId;
}

unsigned int RenderPrimitive::getIndicesBuffer()
{
    return this->mIndicesBuffer;
}

void RenderPrimitive::setIndicesBuffer(unsigned int bufferId)
{
    this->mIndicesBuffer = bufferId;
}

void RenderPrimitive::setMatrix(glm::mat4* matrix)
{
    this->pMatrix = matrix;
}

glm::mat4* RenderPrimitive::getMatrix()
{
    return this->pMatrix;
}

void RenderPrimitive::setCurrentMatrix(const glm::mat4& matrix)
{
    mCurrentMatrix = matrix;
}

glm::mat4 RenderPrimitive::getCurrentMatrix()
{
    return mCurrentMatrix;
}

void RenderPrimitive::setParentMatrixes(vector<glm::mat4>* parentMatrixes)
{
    for (unsigned int i = 0; i < parentMatrixes->size(); i++)
    {
        mParentMatrixes.push_back((*parentMatrixes)[i]);
    }
}

vector<glm::mat4>* RenderPrimitive::getParentMatrixes()
{
    return& mParentMatrixes;
}



GLuint RenderWithFBO::mFboProId = 0;

GLuint RenderWithFBO::mFboId = 0;
GLuint RenderWithFBO::mRboId = 0;
GLuint RenderWithFBO::mTexId = 0;

GLuint RenderWithFBO::mInverseFboId = 0;
GLuint RenderWithFBO::mInverseRboId = 0;
GLuint RenderWithFBO::mInverseTexId = 0;

GLuint RenderWithFBO::mVertexBuf = 0;
GLuint RenderWithFBO::mTexCoordBuf = 0;

void RenderWithFBO::inverseBitMap(int width, int height)
{
    GLuint proId = loadFboShader(INVERSEVERT, INVERSEFRAG);
    GLuint texCoordBuf = 0;
    GLuint vertexBuf = 0;
    GLfloat coordVertices[] =
    {
        0.0f, 1.0f,
        1.0f, 1.0f,
        1.0f, 0.0f,
        0.0f, 0.0f,
    };
    GLfloat squareVertices[] =
    {
        -1.0f, -1.0f, -1.0,
        1.0f, -1.0f, -1.0,
        1.0f,  1.0f, -1.0,
        -1.0f,  1.0f, -1.0
    };
    setBufferForFbo(texCoordBuf, vertexBuf, coordVertices,
                    sizeof(coordVertices), squareVertices, sizeof(squareVertices));
    createAndBindInverseFBO(width, height);
    glViewport(0, 0, width, height);
    inverseTexture(proId, texCoordBuf, vertexBuf);
}

GLuint RenderWithFBO::loadFboShader(const char* vShader,
                                    const char* fShader)
{
    ShaderProgram shaderPro;
    GLuint progId = glCreateProgram();
    std::string vertShaderName = vShader;
    std::string fragShaderName = fShader;
    shaderPro.loadShader(progId, vertShaderName, GL_VERTEX_SHADER);
    shaderPro.loadShader(progId, fragShaderName, GL_FRAGMENT_SHADER);
    return progId;
}

void RenderWithFBO::setBufferForFbo(GLuint& texCoordBuf,
                                    GLuint& vertexBuf, GLfloat* pCoord, GLuint numCoord,
                                    GLfloat* pSquare, GLuint numSquare)
{
    glGenBuffers(1, &texCoordBuf);
    glBindBuffer(GL_ARRAY_BUFFER, texCoordBuf);
    glBufferData(GL_ARRAY_BUFFER, numCoord,
                 pCoord, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glGenBuffers(1, &vertexBuf);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuf);
    glBufferData(GL_ARRAY_BUFFER, numSquare,
                 pSquare, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void RenderWithFBO::createAndBindInverseFBO(int width, int height)
{
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    createRenderObj(width, height, mInverseRboId);
    createTextureObj(width, height, mInverseTexId);
    createFrameBufferObj(width, height, mInverseFboId,
                         mInverseTexId, mInverseRboId);
    glBindFramebuffer(GL_FRAMEBUFFER, mInverseFboId);
}

void RenderWithFBO::createAndBindBitMapFBO(int width, int height)
{
    createRenderObj(width, height, mRboId);
    createTextureObj(width, height, mTexId);
    createFrameBufferObj(width, height, mFboId,
        mTexId, mRboId);
    glBindFramebuffer(GL_FRAMEBUFFER, mFboId);
}

void RenderWithFBO::inverseTexture(GLuint proId, GLuint texCoordBuf,
                                   GLuint vertexBuf)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    GLuint texCoordId = glGetAttribLocation(proId, "texCoord");
    GLuint vertexId   = glGetAttribLocation(proId, "vPosition");
    GLuint textureId  = glGetUniformLocation(proId, "RenderTex");

    glUseProgram(proId);
    glEnableVertexAttribArray(vertexId);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuf);
    glVertexAttribPointer(vertexId, 3, GL_FLOAT, GL_FALSE,
                          0, (void*)0);

    glEnableVertexAttribArray(texCoordId);
    glBindBuffer(GL_ARRAY_BUFFER, texCoordBuf);
    glVertexAttribPointer(texCoordId, 2, GL_FLOAT, GL_FALSE,
                          0, (void*)0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, mTexId);
    glUniform1i(textureId, 0);
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableVertexAttribArray(vertexId);
    glDisableVertexAttribArray(texCoordId);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glUseProgram(0);
}

void RenderWithFBO::createRenderObj(int width, int height, GLuint& rboId)
{
    glGenRenderbuffers(1, &rboId);
    glBindRenderbuffer(GL_RENDERBUFFER, rboId);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
}

void RenderWithFBO::createTextureObj(int width, int height, GLuint& texId)
{
    glGenTextures(1, &texId);
    glBindTexture(GL_TEXTURE_2D, texId);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB,
                 GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
}

bool RenderWithFBO::createFrameBufferObj(int width, int height, GLuint& fboId,
        GLuint texId, GLuint rboId)
{
    glGenFramebuffers(1, &fboId);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return false;
    }
    glBindFramebuffer(GL_FRAMEBUFFER, fboId);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, texId);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                           GL_TEXTURE_2D, texId, 0);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return false;
    }
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);

    glBindRenderbuffer(GL_RENDERBUFFER, rboId);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                              GL_RENDERBUFFER, rboId);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return false;
    }
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    return true;
}

int RenderWithFBO::renderFboTexture()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUseProgram(mFboProId);

    GLuint texCoordId = glGetAttribLocation(mFboProId, "texCoord");
    GLuint vertexId   = glGetAttribLocation(mFboProId, "vPosition");
    GLuint textureId  = glGetUniformLocation(mFboProId, "RenderTex");

    glEnableVertexAttribArray(vertexId);
    glBindBuffer(GL_ARRAY_BUFFER, mVertexBuf);
    glVertexAttribPointer(vertexId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);

    glEnableVertexAttribArray(texCoordId);
    glBindBuffer(GL_ARRAY_BUFFER, mTexCoordBuf);
    glVertexAttribPointer(texCoordId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);

    glBindTexture(GL_TEXTURE_2D, mTexId);
    glUniform1i(textureId, 0);
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableVertexAttribArray(texCoordId);
    glDisableVertexAttribArray(vertexId);
    return 0;
}

void RenderWithFBO::createAndBindFbo(int width, int height)
{
    if (0 == mFboId)
    {
        mFboProId = loadFboShader(FBOVERT, FBOFRAG);

        createRenderObj(width, height, mRboId);
        createTextureObj(width, height, mTexId);
        createFrameBufferObj(width, height, mFboId, mTexId, mRboId);

        GLfloat coordVertices[] =
        {
            0.0f, 0.0f,
            1.0f, 0.0f,
            1.0f, 1.0f,
            0.0f, 1.0f,
        };
        GLfloat squareVertices[] =
        {
            -1.0f, -1.0f,
            1.0f, -1.0f,
            1.0f,  1.0f,
            -1.0f,  1.0f,
        };
        setBufferForFbo(mTexCoordBuf, mVertexBuf, coordVertices,
            sizeof(coordVertices), squareVertices, sizeof(squareVertices));
    }
}

void RenderWithFBO::releaseFbo()
{
    if (0 != mFboId)
    {
        glDeleteFramebuffers(1, &mFboId);
        glDeleteRenderbuffers(1, &mRboId);
        glDeleteTextures(1, &mTexId);
        mFboId = 0;
    }
}



CPhysicalCamera* RenderScene::sCamera = NULL;

RenderScene::RenderScene()
    : pScene(0), pLight(0), pCamera(0)
    , mCurrentTime(0), mLastPlaying(0), mDuration(0)
    , mAnimationPlay(false), mAnimationLoop(true)
    , mEnableRotation(false), mRotationDuration(0)
    , mRotationNow(true), mCurrentImage("")
{
#if ENABLE_FPS
    pFPSCounter = NULL;
#endif
}

RenderScene::~RenderScene()
{
#if ENABLE_FPS
    delete pFPSCounter;
#endif
    if (pScene->getLight())
    {
        delete pLight;
    }
    delete pCamera;
    RenderPrimitive* pRenderPrimitive;
    unsigned int size = mPrimitiveVec.size();
    for (unsigned int i = 0; i < size; ++i)
    {
        pRenderPrimitive = mPrimitiveVec[i];
        delete pRenderPrimitive;
    }
    mPrimitiveVec.clear();
    mBindBufferMap.clear();
    delete pScene;
}

Scene* RenderScene::loadScene(const std::string& directory,
                              const std::string& file)
{
    Scene* pScene = new Scene();
    mLoadJson.setJsonInfo(directory, file);
    if (!mLoadJson.parseJsonFile())
    {
        return 0;
    }

    if (!mLoadJson.parseScene(pScene))
    {
        return 0;
    }
    return pScene;
}

Scene* RenderScene::loadScene(glTFHandle* handle)
{
    Scene* pScene = new Scene();
    pScene->setGltfHandle(handle);
    try
    {
        if (!mLoadJson.parseScene(pScene))
        {
            return 0;
        }    
    }
    catch (boost::property_tree::ptree_error& e)
    {
        fprintf(stderr, "%s\n", e.what());
        delete pScene;
        return 0;
    }
    catch (boost::exception const&)
    {
        fprintf(stderr, "%s\n", "unknown boost error");
        delete pScene;
        return 0;
    }
    
    return pScene;
}

glTFHandle* RenderScene::initScene(const glTFFile* jsonfile)
{
    if (0 == jsonfile)
    {
        return 0;
    }
    string tmpStr(jsonfile->buffer, jsonfile->size);
    mLoadJson.setJsonInfo(tmpStr);
    if (!mLoadJson.parseJsonFile())
    {
        return 0;
    }
    string tmpFileName(jsonfile->filename);
    return mLoadJson.getFileNameInJson(tmpFileName);
}

unsigned int RenderScene::bindAttribute(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ARRAY_BUFFER,
                 pAttr->getDataCount() * pAttr->getByteStride(),
                 pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

unsigned int RenderScene::bindIndices(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                 pAttr->getDataCount() * pAttr->getByteStride(),
                 pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

void RenderScene::constructMesh(const std::string meshName,
                                glm::mat4* pmatrix,
                                Node* pNode)
{
    const Mesh* pMesh = pScene->findMesh(meshName);
    unsigned int size = pMesh->getPrimitiveVecSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        constructPrimitive(pMesh->getPrimitiveVec(i), pmatrix, pNode);
    }
}

void RenderScene::constructPrimitive(const Primitives* pPrimitive,
                                     glm::mat4* pmatrix,
                                     Node* pNode)
{
    Material* pMaterial = pScene->findMaterial(
                            pPrimitive->getMaterialIndex());
    if (0 == pMaterial)
    {
        return;
    }
    RenderPrimitive* pRenderPrimitive = new RenderPrimitive();
    bindAttributeBuffer(pPrimitive, pRenderPrimitive);
    pRenderPrimitive->setMaterial(pMaterial);
    pRenderPrimitive->setCurrentMatrix(pNode->getCurrentNodeMatrix());
    pRenderPrimitive->setParentMatrixes(pNode->getParentMatrixes());
    pRenderPrimitive->setMatrix(pmatrix);

    pushPrimitiveVec(pRenderPrimitive);
}

void RenderScene::bindAttributeBuffer(const Primitives* pPrimitive,
                                      RenderPrimitive* pRP)
{
    const Attribute* pAttr;
    BindBufferInfo bindBufferInfo;
    std::map<std::string, BindBufferInfo>::const_iterator it;

    const std::string& posId = pPrimitive->getAttributeIndex("POSITION");
    it = mBindBufferMap.find(posId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(posId);
        if (0 != pAttr)
        {
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            bindBufferInfo.mDataCount = pAttr->getDataCount();
            pRP->setVertexBuffer(bindBufferInfo.mBufferId);
            pRP->setVerterCount(bindBufferInfo.mDataCount);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (posId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setVertexBuffer(it->second.mBufferId);
        pRP->setVerterCount(it->second.mDataCount);
    }

    const std::string& NorId = pPrimitive->getAttributeIndex("NORMAL");
    it = mBindBufferMap.find(NorId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(NorId);
        if (0 != pAttr)
        {
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            pRP->setNormalBuffer(bindBufferInfo.mBufferId);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (NorId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setNormalBuffer(it->second.mBufferId);
    }

    const std::string& TexId = pPrimitive->getAttributeIndex("TEXCOORD_0");
    it = mBindBufferMap.find(TexId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(TexId);
        if (0 != pAttr)
        {
            float* pData = (float*)pAttr->getAttributeData();
            unsigned int size = pAttr->getDataCount();
            for (unsigned int i = 0; i < size; i++)
            {
                pData[i*2+1] = 1 - pData[i*2+1];
            }
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            pRP->setTexCoordBuffer(bindBufferInfo.mBufferId);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (TexId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setTexCoordBuffer(it->second.mBufferId);
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("JOINT"));
    if (0 != pAttr)
    {
        pRP->setJointBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("WEIGHT"));
    if (0 != pAttr)
    {
        pRP->setWeightBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getIndicesIndex());
    if (0 != pAttr)
    {
        pRP->setIndicesBuffer(bindIndices(pAttr));
        pRP->setIndicesCount(pAttr->getDataCount());
        pRP->setIndicesDataType(pAttr->getDataType());
    }
}

void RenderScene::pushPrimitiveVec(RenderPrimitive* pPrim)
{
    (this->mPrimitiveVec).push_back(pPrim);
}

RenderPrimitive* RenderScene::getPrimitiveVec(unsigned int index)
{
    if (this->mPrimitiveVec.size() <= index)
    {
        return 0;
    }
    return this->mPrimitiveVec[index];
}

unsigned int RenderScene::getPrimitiveVecSize()
{
    return this->mPrimitiveVec.size();
}

glm::mat4* RenderScene::calcProjectionMatrix()
{
    const Node* pParentNode = pScene->getRootNode();
    glm::mat4 projection;
    if (!pParentNode->getFlagCamera())
    {
        projection = glm::perspective(37.8492f, 1.5f, 1.0f, 500000.0f);
    }
    else
    {
        ParseCamera* cCamera  = pScene->findCamera(pScene->getCameraIndex());
        projection = glm::perspective(cCamera->getYFov(),
                                      cCamera->getAspectRatio(),
                                      cCamera->getNear(),
                                      cCamera->getFar());
    }
    glm::mat4* pProject = new glm::mat4;

    memcpy(pProject, glm::value_ptr(projection), sizeof(glm::mat4));
    return pProject;
}

void RenderScene::initOpengl()
{
    glClearColor(0.5f, 0.5f, 0.5f, 0.5f);
    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glDepthFunc(GL_LESS);
}

void RenderScene::createDefaultLight()
{
    if (pScene->getLight())
    {
        pLight = new Light();
        light0Transfrom = new glm::mat4;
        memset((void*)light0Transfrom, 0, sizeof(glm::mat4));
    }
    else
    {
        pLight = pScene->findLight(pScene->GetLightIndex());
        light0Transfrom = pScene->findLightTransformation();
    }
    if (pLight->getPosFlag())
    {
        glLightfv(GL_LIGHT0, GL_POSITION, (GLfloat*)&pLight->getPosition());
    }
    else
    {
        glLightfv(GL_LIGHT0, GL_POSITION, pLight->getPositions());
    }
    glLightfv(GL_LIGHT0, GL_AMBIENT, pLight->getColorAmbient());
    glLightfv(GL_LIGHT0, GL_DIFFUSE, pLight->getColorDiffuse());
    glLightfv(GL_LIGHT0, GL_SPECULAR, pLight->getColorSpecular());
    glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 45.0);
}

void RenderScene::createDefaultCamera()
{
    glm::vec3 vDelta = pScene->getVertexMax() - pScene->getVertexMin();
    glm::vec3 vHalf =  pScene->getVertexMin() + (vDelta / 2.0f);

    float fPos = sqrt(vDelta.x * vDelta.x
                      + vDelta.y * vDelta.y
                      + vDelta.z * vDelta.z);
    if (fPos >= 3000)
    {
        pCamera = new CPhysicalCamera(
            glm::vec3(vHalf.x, -fPos * 0.6f, fPos * 0.4f),
            vHalf,
            glm::vec3(0.0f, 0.0f, 1.0f));
    }
    else
    {
        pCamera = new CPhysicalCamera(
            glm::vec3(vHalf.x, -fPos, 2 * fPos),
            vHalf,
            glm::vec3(.0f, 0.0f, 1.0f));
    }
    pCamera->flength = fPos;
    pCamera->vModelCenterPos = vHalf;
    pCamera->SetRotationAngle();
    if (sCamera == NULL)
    {
        sCamera = pCamera;
    }
    else if (sCamera->flength <= pCamera->flength)
    {
        sCamera = pCamera;
    }
}

int RenderScene::initRender(glTFHandle* handle)
{
    pScene = loadScene(handle);
    if (0 == pScene)
    {
        return -1;
    }

    initOpengl();
    createDefaultLight();
    createDefaultCamera();

    Node* pNode;
    unsigned int nSize = pScene->getNodeSize();
    for (unsigned int i = 0; i < nSize; ++i)
    {
        pNode = pScene->getNode(i);
        if (!pNode->hasMesh())
        {
            continue;
        }
        unsigned int mSize = pNode->getMeshIndexSize();
        for (unsigned int j = 0; j < mSize; ++j)
        {
            constructMesh(pNode->getMeshIndex(j), pNode->getMatrix(),
                          pNode);
        }
    }

    if (0 != pScene->getAnimationCount())
    {
        Animation* pAnimation = pScene->getAnimation(0);
        this->mDuration = pAnimation->getDuration();
    }

    if (pScene->getSkin())
    {
        updateBoneMatrix(pScene->getRootBone(), glm::mat4(1.0f));
    }

#if ENABLE_FPS
    pFPSCounter = new FPSCounter();
    pFPSCounter->loadFPSShader(&mShaderProgram);
#endif
    return 0;
}

void RenderScene::renderPrimitive(RenderPrimitive* pPrimitive,
                                  unsigned int progId)
{
    upLoadMatrixInfo(progId, pPrimitive);
    upLoadUniform(progId, pPrimitive);
    upLoadAttribute(progId, pPrimitive);

#if ENABLE_FPS
    //pFPSCounter->gpuTimeStampStart();
    drawTriangle(pPrimitive);
    //pFPSCounter->gpuTimeStampEnd();
    //pFPSCounter->gpuTimeAccumulation();
#else
    drawTriangle(pPrimitive);
#endif
}

void RenderScene::upLoadMatrixInfo(unsigned int progId,
                                   RenderPrimitive* pPrimitive)
{
    glm::mat4* pMatrix = pPrimitive->getMatrix();
    vector<glm::mat4>* parentMatrixes = pPrimitive->getParentMatrixes();

    if (mEnableRotation && parentMatrixes->size() > 0 && mRotationNow)
    {
        glm::mat4 rootMatrix = (*parentMatrixes)[parentMatrixes->size() - 1];
        glm::mat4 tmpMat4;
        tmpMat4 = glm::rotate(rootMatrix, 0.3f, glm::vec3(0.0f, 0.0f, 1.0f));
        (*parentMatrixes)[parentMatrixes->size() - 1] = tmpMat4;
        for (int i = (int)parentMatrixes->size() - 2; i >= 0; i--)
        {
            tmpMat4 = tmpMat4 * (*parentMatrixes)[i];
        }
        tmpMat4 = tmpMat4 * pPrimitive->getCurrentMatrix();
        *pMatrix = tmpMat4;
    }
    glm::mat4 modelViewMatrix = sCamera->getModelViewMatrix(pMatrix);
    mShaderProgram.setUniform(progId, "u_modelViewMatrix", modelViewMatrix);

    glm::mat3 normalMatrix(*pMatrix);
    glm::mat3 NormalInverseTranspose = glm::mat3(sCamera->Look()) *
                                       (glm::transpose(glm::inverse(
                                               normalMatrix)));
    mShaderProgram.setUniform(progId, "u_normalMatrix",
                              NormalInverseTranspose);

    glm::mat4* pProject = calcProjectionMatrix();
    mShaderProgram.setUniform(progId, "u_projectionMatrix", pProject);
    delete pProject;

    mShaderProgram.setUniform(progId, "M", (*pMatrix));
    mShaderProgram.setUniform(progId, "V", (sCamera->Look()));
}

bool RenderScene::upLoadTechProperty()
{
    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_TRUE);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    return true;
}

void RenderScene::upLoadTechInfo(unsigned int progId)
{
    glm::vec3 color = pLight->getColor();
    mShaderProgram.setUniform(progId, "u_light0Color", &color);
    mShaderProgram.setUniform(progId, "u_light0Transform", light0Transfrom);
}

void RenderScene::upLoadAnimation(unsigned int progId)
{
    if (!this->mAnimationLoop &&
        (fmod(mCurrentTime, mDuration) < fmod(mLastPlaying, mDuration)))
    {
        this->mAnimationPlay = false;
    }

    if (this->mAnimationPlay)
    {
        updateBoneMatrix(pScene->getRootBone(), glm::mat4(1.0f));
    }

    Skin* pSkin = pScene->getSkin();
    glm::mat4* pBindMatrix = pSkin->getBindMatrix();
    glm::mat4* pTempMatrix = new glm::mat4[pSkin->getBindMatrixCount()];
    memcpy(pTempMatrix, pBindMatrix,
           sizeof(glm::mat4)*pSkin->getBindMatrixCount());

    unsigned int jointCount = pSkin->getBoneIdSize();
    for (unsigned int i = 0; i < jointCount; i++)
    {
        Bone* pBone = pScene->findBone(pSkin->getBoneId(i));
        pTempMatrix[i] = pBone->getGlobleMatrix() * pTempMatrix[i];
    }

    int iLoc = glGetUniformLocation(progId, "u_jointMat");
    glUniformMatrix4fv(iLoc, pSkin->getBindMatrixCount(), false,
                       (GLfloat*)pTempMatrix);
    delete[] pTempMatrix;
}

void RenderScene::upLoadUniform(unsigned int progId,
                                RenderPrimitive* pPrimitive)
{
    int textureId = 0;
    Material* pMaterial = pPrimitive->getMaterial();
    unsigned int size = pMaterial->getMaterialProperSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        const MaterialProperty* pProper = pMaterial->getMaterialProper(i);
        const std::string& properName = pProper->getPropertyName();
        switch (pProper->getDataType())
        {
            case DataType_FLOAT:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (float*)(pProper->
                                                   getPropertyData()));
                break;
            case DataType_FLOAT_VEC2:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::vec2*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_VEC3:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::vec3*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_VEC4:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::vec4*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_MAT3:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::mat3*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_MAT4:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::mat4*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_SAMPLER_2D:
                {
#if LOAD_ONCE
                    if (mCurrentImage == pProper->getImagePath())
                    {
                        break;
                    }
#endif 
                    mShaderProgram.setUniform(progId, properName.c_str(),
                        textureId);
                    Texture* oneTexture =
                        pScene->findTexture(pProper->getImagePath());
                    mCurrentImage = pProper->getImagePath();
                    oneTexture->bindTexture(textureId);
                    textureId++;
                }
                break;
            default:
                break;
        }
    }
    return;
}

void RenderScene::upLoadAttribute(unsigned int progId,
                                  RenderPrimitive* pPrimitive)
{
    int tmpId = glGetAttribLocation(progId, "a_position");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getVertexBuffer());
        glVertexAttribPointer(tmpId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_normal");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getNormalBuffer());
        glVertexAttribPointer(tmpId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_texcoord0");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getTexCoordBuffer());
        glVertexAttribPointer(tmpId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_weight");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getWeightBuffer());
        glVertexAttribPointer(tmpId, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_joint");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getJointBuffer());
        glVertexAttribPointer(tmpId, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    return;
}

void RenderScene::drawTriangle(RenderPrimitive* pPrimitive)
{
    if (pPrimitive->getIndicesCount() > 0)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, pPrimitive->getIndicesBuffer());
        glDrawElements(GL_TRIANGLES, pPrimitive->getIndicesCount(),
                       pPrimitive->getIndicesDataType(), 0);
    }
    else
    {
        glDrawArrays(GL_TRIANGLES, 0, pPrimitive->getVerterCount());
    }
}

void RenderScene::prepareRender(void* lpParam)
{
#if ENABLE_FPS
    pFPSCounter->timeStamp();
    pFPSCounter->gpuAccumulationTimeclear();
#endif
    viewport = *((glTFViewport*)lpParam);
    RenderWithFBO::createAndBindFbo(viewport.width * SSAA, viewport.height * SSAA);
    glBindFramebuffer(GL_FRAMEBUFFER, RenderWithFBO::mFboId);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(viewport.x, viewport.y, viewport.width * SSAA, viewport.height * SSAA);
    return;
}

void RenderScene::render()
{
    mCurrentTime += clock() / double(CLOCKS_PER_SEC) - mLastPlaying;
    mRotationDuration += mCurrentTime;
    if (mRotationDuration > ((double)CLOCKS_PER_SEC / 10))
    {
        mRotationNow = true;
        mRotationDuration = 0;
    }
    else
    {
        mRotationNow = false;
    }

    upLoadTechProperty();
    unsigned int size = pScene->getTechSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        renderShader(pScene->getTechnique(i));
    }
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
    mLastPlaying = mCurrentTime;
}

void RenderScene::renderShader(Technique* pTechnique)
{
    if (!pTechnique->getProgramState())
    {
        return;
    }

    pTechnique->useTechnique();
    unsigned int progId = pTechnique->getProgramId();
    upLoadTechInfo(progId);

    if (pScene->hasSkin())
    {
        upLoadAnimation(progId);
    }

    RenderPrimitive* pPrimitive;
    const Material* pMaterial;
    const std::string& techId = pTechnique->getTechId();
    unsigned int size = getPrimitiveVecSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        pPrimitive = getPrimitiveVec(i);
        pMaterial = pPrimitive->getMaterial();
        if (pMaterial->getTechniqueId() == techId)
        {
            renderPrimitive(pPrimitive, progId);
        }
    }
}

void RenderScene::completeRender()
{
#if ENABLE_FPS
    pFPSCounter->printFPS(&viewport);
#endif
    glViewport(viewport.x, viewport.y, viewport.width, viewport.height);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    RenderWithFBO::renderFboTexture();
    return;
}

void RenderScene::releaseRender(void* pParam)
{
    mLoadJson.releaseFileName();
    RenderWithFBO::releaseFbo();
    return;
}

void RenderScene::updateJointInfo(Bone* pBone)
{
    Animation* pAnimation;
    AnimChannel* pChannel;
    const std::string& boneId = pBone->getJointId();
    double time = fmod(this->mCurrentTime, this->mDuration);
    unsigned int count = pScene->getAnimationCount();
    for (unsigned int i = 0; i < count; ++i)
    {
        pAnimation = pScene->getAnimation(i);
        if (boneId != pAnimation->getBoneId())
            continue;

        pChannel = pAnimation->getRotateChl();
        if (0 != pChannel)
        {
            glm::vec4* pRotate = pChannel->getTimeValue4(time);
            pBone->setRotate(pRotate);
        }

        pChannel = pAnimation->getTransChl();
        if (0 != pChannel)
        {
            glm::vec3* pTrans = pChannel->getTimeValue3(time);
            pBone->setTranslate(pTrans);
        }

        pChannel = pAnimation->getScaleChl();
        if (0 != pChannel)
        {
            glm::vec3* pScale = pChannel->getTimeValue3(time);
            pBone->setScale(pScale);
        }

        break;
    }
    return;
}
void RenderScene::updateBoneMatrix(Bone* pBone, const glm::mat4& matrix)
{
    if (pBone->getJointFlag())
    {
        updateJointInfo(pBone);
        glm::mat4 scaleMatrix = glm::scale(pBone->getScale());
        glm::mat4 transMatrix = glm::translate(pBone->getTranslate());
        glm::vec4 rotate = pBone->getRotate();
        glm::mat4 rotateMatrix = glm::rotate(transMatrix,
            rotate.w*57, glm::vec3(rotate));
        glm::mat4 localMatrix = rotateMatrix * scaleMatrix;
        pBone->setLocalMatrix(&localMatrix);
    }

    glm::mat4 globleMatrix = matrix * pBone->getLocalMatrix();
    pBone->setGlobleMatrix(&globleMatrix);

    if (!pBone->hasChildren())
    {
        return;
    }

    unsigned int count = pBone->getChildBoneSize();
    for (unsigned int i = 0; i < count; ++i)
    {
        updateBoneMatrix(pBone->getChildBone(i), globleMatrix);
    }

    return;
}

CPhysicalCamera* RenderScene::getCamera()
{
    return sCamera;
}

void RenderScene::prepareRenderBitmap(glTFViewport* pViewport)
{
    RenderWithFBO::createAndBindBitMapFBO(pViewport->width, pViewport->height);
    glViewport(pViewport->x, pViewport->y, pViewport->width, pViewport->height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
}

void RenderScene::renderBitmap()
{
    render();
}

void RenderScene::completeRenderBitmap(glTFViewport* pViewport,
                                       unsigned char* buffer, GLenum format)
{
    glDisable(GL_DEPTH_TEST);
    RenderWithFBO::inverseBitMap(pViewport->width, pViewport->height);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return;
    }
    glReadPixels(0, 0, pViewport->width, pViewport->height, format,
                 GL_UNSIGNED_BYTE, buffer);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glDeleteFramebuffers(1, &RenderWithFBO::mInverseFboId);
    glDeleteRenderbuffers(1, &RenderWithFBO::mInverseRboId);
    glDeleteTextures(1, &RenderWithFBO::mInverseTexId);
}

void RenderScene::getCameraPos(glm::vec3* Eye, glm::vec3* view,
                               glm::vec3* up)
{
    *Eye =  sCamera->vEye;
    *view = sCamera->vView;
    *up = sCamera->vUp;
}

glm::vec3* RenderScene::getModelCenterPos()
{
    return &sCamera->vModelCenterPos;
}

void RenderScene::renderRotateCamera(double horizontal, double vertical,
                                     double planar, double time)
{
    sCamera->rotateCamera(horizontal, vertical, planar, time);
}

void RenderScene::renderMoveCamera(double x, double y, double z,
                                   double time)
{
    sCamera->moveCamera(x, y, z, time);
}

void RenderScene::EnableRotation()
{
    mEnableRotation = true;
}

void RenderScene::DisableRotation()
{
    mEnableRotation = false;
}

double RenderScene::getModelSize()
{
    return sCamera->flength;
}

void RenderScene::startAnimation()
{
    this->mAnimationPlay = true;
}

void RenderScene::stopAnimation()
{
    this->mAnimationPlay = false;
}

void RenderScene::setAnimTime(double time)
{
    this->mLastPlaying = clock() / double(CLOCKS_PER_SEC) - time;
}

double RenderScene::getAnimTime()
{
    errno = 0;
    double time = fmod(this->mCurrentTime, this->mDuration);
    return errno == EDOM ? 0.0 : time;
}

void RenderScene::setAnimLoop(int loop)
{
    if (0 != loop)
    {
        this->mAnimationLoop = true;
    }
    else
    {
        this->mAnimationLoop = false;
    }
}

int  RenderScene::getAnimLoop()
{
    return this->mAnimationLoop ? 1 : 0;
}

double RenderScene::getAnimDuration()
{
    return this->mDuration;
}

int  RenderScene::isAnimPlay()
{
    return this->mAnimationPlay ? 1 : 0;
}