/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#ifndef RENDERSCENE_H
#define RENDERSCENE_H

#include "Shaders.h"
#include "Common.h"
#include "Texture.h"
#include "Camera.h"
#include "LoadScene.h"
#include "FPSCounter.h"

using namespace glTF;

#define FBOVERT "Resource/FboRenderTexture.vert"
#define FBOFRAG "Resource/FboRenderTexture.frag"

#define INVERSEVERT "Resource/inverse.vert"
#define INVERSEFRAG "Resource/inverse.frag"
#define LOAD_ONCE 0
#define SSAA 2
class RenderPrimitive
{
public:
    class Material* getMaterial();
    void setMaterial(class Material* pMaterial);

    unsigned int getVerterCount();
    void setVerterCount(unsigned int count);

    unsigned int getIndicesCount();
    void setIndicesCount(unsigned int count);

    unsigned int getVertexBuffer();
    void setVertexBuffer(unsigned int bufferId);

    unsigned int getNormalBuffer();
    void setNormalBuffer(unsigned int bufferId);

    unsigned int getTexCoordBuffer();
    void setTexCoordBuffer(unsigned int bufferId);

    unsigned int getJointBuffer();
    void setJointBuffer(unsigned int bufferId);

    unsigned int getWeightBuffer();
    void setWeightBuffer(unsigned int bufferId);

    unsigned int getIndicesBuffer();
    void setIndicesBuffer(unsigned int bufferId);

    DataType_E getIndicesDataType();
    void setIndicesDataType(DataType_E type);

    void setMatrix(glm::mat4* pMatrix);
    glm::mat4* getMatrix();

    void setCurrentMatrix(const glm::mat4& currentMatrix);
    glm::mat4 getCurrentMatrix();

    void setParentMatrixes(vector<glm::mat4>* parentMatrixes);
    vector<glm::mat4>* getParentMatrixes();

    RenderPrimitive();
    ~RenderPrimitive();
private:
    DataType_E mIndicesDataType;
    glm::mat4* pMatrix;
    glm::mat4 mCurrentMatrix;
    vector<glm::mat4> mParentMatrixes;
    class Material* pMaterial;
    unsigned int mVerterCount;
    unsigned int mIndicesCount;
    unsigned int mVertexBuffer;
    unsigned int mNormalBuffer;
    unsigned int mTexCoordBuffer;
    unsigned int mJointBuffer;
    unsigned int mWeightBuffer;
    unsigned int mIndicesBuffer;
};

class RenderWithFBO
{
public:
    static void createAndBindFbo(int width, int height);
    static void releaseFbo();
    static int renderFboTexture();

    static void createTextureObj(int width, int height, GLuint& texId);
    static void createRenderObj(int width, int height, GLuint& RboId);
    static bool createFrameBufferObj(int width, int height, GLuint& FboId,
                                     GLuint texId, GLuint RboId);
    static GLuint loadFboShader(const char* vShader, const char* fShader);
    static void setBufferForFbo(GLuint& texCoordBuf, GLuint& vertexBuf,
                                GLfloat* pCoord, GLuint numCoord,
                                GLfloat* pSquare, GLuint numSquare);
    static void createAndBindInverseFBO(int width, int height);
    static void createAndBindBitMapFBO(int width, int height);
    static void inverseBitMap(int width, int height);
    static void inverseTexture(GLuint proId,
                               GLuint texCoordBuf, GLuint vertexBuf);

    static GLuint mFboProId;

    static GLuint mFboId;
    static GLuint mRboId;
    static GLuint mTexId;

    static GLuint mInverseFboId;
    static GLuint mInverseRboId;
    static GLuint mInverseTexId;

    static GLuint mVertexBuf;
    static GLuint mTexCoordBuf;
};

class RenderScene
{
    struct BindBufferInfo
    {
        unsigned int mBufferId;
        unsigned int mDataCount;
    };
public:
    // For camera
    glm::vec3* getModelCenterPos();
    void getCameraPos(glm::vec3* Eye, glm::vec3* view, glm::vec3* up);
    double getModelSize();
    CPhysicalCamera* getCamera();
    void renderRotateCamera(double horizontal, double vertical,
                            double planar, double time);
    void renderMoveCamera(double x, double y, double z, double time);

    // For render scene
    glTFHandle* initScene(const glTFFile* jsonfile);
    int  initRender(glTFHandle* handle);
    void prepareRender(void* lpParam);
    void render();
    void completeRender();
    void releaseRender(void* lpParam);

    // For bitmap
    static void prepareRenderBitmap(glTFViewport* pViewport);
    void renderBitmap();
    static void completeRenderBitmap(glTFViewport* pViewport,
                                     unsigned char* buffer, GLenum format);
    // For animation
    void startAnimation();
    void stopAnimation();
    void setAnimTime(double time);
    double getAnimTime();
    void setAnimLoop(int loop);
    int  getAnimLoop();
    double getAnimDuration();
    int  isAnimPlay();

    // For model rotation
    void EnableRotation();
    void DisableRotation();

    RenderScene();
    ~RenderScene();

private:
    void initOpengl();
    class Scene* loadScene(const std::string& directory,
                                   const std::string& file);
    class Scene* loadScene(glTFHandle* handle);
    void createDefaultLight();
    void createDefaultCamera();
    void constructMesh(const std::string meshName, glm::mat4* pmatrix,
                       Node* pNode);
    void constructPrimitive(const Primitives* pPrim, glm::mat4* pmatrix,
                            Node* pNode);
    void pushPrimitiveVec(RenderPrimitive* pPrim);
    class RenderPrimitive* getPrimitiveVec(unsigned int index);
    unsigned int getPrimitiveVecSize();
    unsigned int bindAttribute(const Attribute* pAttr);
    unsigned int bindIndices(const Attribute* pAttr);
    void bindAttributeBuffer(const Primitives* pPrimitive,
                             RenderPrimitive* pRP);
    glm::mat4* calcProjectionMatrix();
    bool upLoadTechProperty();
    void upLoadTechInfo(unsigned int progId);
    void upLoadMatrixInfo(unsigned int progId, RenderPrimitive* pPrimitive);
    void upLoadAnimation(unsigned int progId);
    void upLoadUniform(unsigned int progId, RenderPrimitive* pPrimitive);
    void upLoadAttribute(unsigned int progId, RenderPrimitive* pPrimitive);
    void updateJointInfo(Bone* pBone);
    void updateBoneMatrix(Bone* pBone, const glm::mat4& matrix);
    void drawTriangle(RenderPrimitive* pPrimitive);
    void renderPrimitive(RenderPrimitive* pPrimitive,
                         unsigned int progId);
    void renderShader(Technique* pTechnique);

    CPhysicalCamera* pCamera;
    static CPhysicalCamera* sCamera;
    Light* pLight;
    glm::mat4* light0Transfrom;

    bool mAnimationPlay;
    bool mAnimationLoop;
    double mCurrentTime;
    double mLastPlaying;
    double mDuration;

    std::vector<RenderPrimitive*> mPrimitiveVec;
    Scene* pScene;
    Parser mLoadJson;
    std::map<std::string, BindBufferInfo> mBindBufferMap;

    ShaderProgram mShaderProgram;
    std::string mCurrentImage;
    glTFViewport viewport;

    bool mEnableRotation;
    bool mRotationNow;
    double mRotationDuration;
    
    bool loadFPSShader();
    int getFPS();
    void printFPS(glTFViewport* pViewpoit);
#if ENABLE_FPS
    FPSCounter* pFPSCounter;
#endif
};

#endif
